"""
Testing puzzles.

"""

# Test of the Puzzles module

import itertools
import re

from mock import Mock

from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils import search
from evennia.utils.create import create_object

from . import puzzles


class TestPuzzles(BaseEvenniaCommandTest):
    def setUp(self):
        super().setUp()
        self.steel = create_object(self.object_typeclass, key="steel", location=self.char1.location)
        self.flint = create_object(self.object_typeclass, key="flint", location=self.char1.location)
        self.fire = create_object(self.object_typeclass, key="fire", location=self.char1.location)
        self.steel.tags.add("tag-steel")
        self.steel.tags.add("tag-steel", category="tagcat")
        self.flint.tags.add("tag-flint")
        self.flint.tags.add("tag-flint", category="tagcat")
        self.fire.tags.add("tag-fire")
        self.fire.tags.add("tag-fire", category="tagcat")

    def _assert_msg_matched(self, msg, regexs, re_flags=0):
        matches = []
        for regex in regexs:
            m = re.search(regex, msg, re_flags)
            self.assertIsNotNone(m, "%r didn't match %r" % (regex, msg))
            matches.append(m)
        return matches

    def _assert_recipe(self, name, parts, results, and_destroy_it=True, expected_count=1):
        def _keys(items):
            return [item["key"] for item in items]

        recipes = search.search_script_tag("", category=puzzles._PUZZLES_TAG_CATEGORY)
        self.assertEqual(expected_count, len(recipes))
        self.assertEqual(name, recipes[expected_count - 1].db.puzzle_name)
        self.assertEqual(parts, _keys(recipes[expected_count - 1].db.parts))
        self.assertEqual(results, _keys(recipes[expected_count - 1].db.results))
        self.assertEqual(
            puzzles._PUZZLES_TAG_RECIPE,
            recipes[expected_count - 1].tags.get(category=puzzles._PUZZLES_TAG_CATEGORY),
        )
        recipe_dbref = recipes[expected_count - 1].dbref
        if and_destroy_it:
            recipes[expected_count - 1].delete()
        return recipe_dbref if not and_destroy_it else None

    def _assert_no_recipes(self):
        self.assertEqual(
            0, len(search.search_script_tag("", category=puzzles._PUZZLES_TAG_CATEGORY))
        )

    # good recipes
    def _good_recipe(self, name, parts, results, and_destroy_it=True, expected_count=1):
        regexs = []
        for p in parts:
            regexs.append(r"^Part %s\(#\d+\)$" % (p))
        for r in results:
            regexs.append(r"^Result %s\(#\d+\)$" % (r))
        regexs.append(r"^Puzzle '%s' %s\(#\d+\) has been created successfully.$" % (name, name))
        lhs = [name] + parts
        cmdstr = ",".join(lhs) + "=" + ",".join(results)
        msg = self.call(puzzles.CmdCreatePuzzleRecipe(), cmdstr, caller=self.char1)
        recipe_dbref = self._assert_recipe(name, parts, results, and_destroy_it, expected_count)
        self._assert_msg_matched(msg, regexs, re_flags=re.MULTILINE | re.DOTALL)
        return recipe_dbref

    def _check_room_contents(self, expected, check_test_tags=False):
        by_obj_key = lambda o: o.key
        room1_contents = sorted(self.room1.contents, key=by_obj_key)
        for key, grp in itertools.groupby(room1_contents, by_obj_key):
            if key in expected:
                grp = list(grp)
                self.assertEqual(
                    expected[key],
                    len(grp),
                    "Expected %d but got %d for %s" % (expected[key], len(grp), key),
                )
                if check_test_tags:
                    for gi in grp:
                        tags = gi.tags.all(return_key_and_category=True)
                        self.assertIn(("tag-" + gi.key, "tagcat"), tags)

    def _arm(self, recipe_dbref, name, parts):
        regexs = [
            r"^Puzzle Recipe %s\(#\d+\) '%s' found.$" % (name, name),
            r"^Spawning %d parts ...$" % (len(parts)),
        ]
        for p in parts:
            regexs.append(r"^Part %s\(#\d+\) spawned .*$" % (p))
        regexs.append(r"^Puzzle armed successfully.$")
        msg = self.call(puzzles.CmdArmPuzzle(), recipe_dbref, caller=self.char1)
        self._assert_msg_matched(msg, regexs, re_flags=re.MULTILINE | re.DOTALL)

    def test_cmdset_puzzle(self):
        self.char1.cmdset.add("evennia.contrib.game_systems.puzzles.PuzzleSystemCmdSet")
        # FIXME: testing nothing, this is just to bump up coverage

    def test_cmd_puzzle(self):
        self._assert_no_recipes()

        # bad syntax
        def _bad_syntax(cmdstr):
            self.call(
                puzzles.CmdCreatePuzzleRecipe(),
                cmdstr,
                "Usage: @puzzle name,<part1[,...]> = <result1[,...]>",
                caller=self.char1,
            )

        _bad_syntax("")
        _bad_syntax("=")
        _bad_syntax("nothing =")
        _bad_syntax("= nothing")
        _bad_syntax("nothing")
        _bad_syntax(",nothing")
        _bad_syntax("name, nothing")
        _bad_syntax("name, nothing =")

        self._assert_no_recipes()

        self._good_recipe("makefire", ["steel", "flint"], ["fire", "steel", "flint"])
        self._good_recipe("hot steels", ["steel", "fire"], ["steel", "fire"])
        self._good_recipe(
            "furnace",
            ["steel", "steel", "fire"],
            ["steel", "steel", "fire", "fire", "fire", "fire"],
        )

        # bad recipes
        def _bad_recipe(name, parts, results, fail_regex):
            cmdstr = ",".join([name] + parts) + "=" + ",".join(results)
            msg = self.call(puzzles.CmdCreatePuzzleRecipe(), cmdstr, caller=self.char1)
            self._assert_no_recipes()
            self.assertIsNotNone(re.match(fail_regex, msg), msg)

        _bad_recipe("name", ["nothing"], ["neither"], r"Could not find 'nothing'.")
        _bad_recipe("name", ["steel"], ["nothing"], r"Could not find 'nothing'.")
        _bad_recipe("", ["steel", "fire"], ["steel", "fire"], r"^Invalid puzzle name ''.")
        self.steel.location = self.char1
        _bad_recipe("name", ["steel"], ["fire"], r"^Invalid location for steel$")
        _bad_recipe("name", ["flint"], ["steel"], r"^Invalid location for steel$")
        _bad_recipe("name", ["self"], ["fire"], r"^Invalid typeclass for Char$")
        _bad_recipe("name", ["here"], ["fire"], r"^Invalid typeclass for Room$")

        self._assert_no_recipes()

    def test_cmd_armpuzzle(self):
        # bad arms
        self.call(
            puzzles.CmdArmPuzzle(),
            "1",
            "A puzzle recipe's #dbref must be specified",
            caller=self.char1,
        )
        self.call(puzzles.CmdArmPuzzle(), "#1", "Invalid puzzle '#1'", caller=self.char1)

        recipe_dbref = self._good_recipe(
            "makefire", ["steel", "flint"], ["fire", "steel", "flint"], and_destroy_it=False
        )

        # delete proto parts and proto result
        self.steel.delete()
        self.flint.delete()
        self.fire.delete()

        # good arm
        self._arm(recipe_dbref, "makefire", ["steel", "flint"])
        self._check_room_contents({"steel": 1, "flint": 1}, check_test_tags=True)

    def _use(self, cmdstr, expmsg):
        msg = self.call(puzzles.CmdUsePuzzleParts(), cmdstr, expmsg, caller=self.char1)
        return msg

    def test_cmd_use(self):
        self._use("", "Use what?")
        self._use("something", "There is no something around.")
        self._use("steel", "You have no idea how this can be used")
        self._use("steel flint", "There is no steel flint around.")
        self._use("steel, flint", "You have no idea how these can be used")

        recipe_dbref = self._good_recipe(
            "makefire", ["steel", "flint"], ["fire"], and_destroy_it=False
        )
        recipe2_dbref = self._good_recipe(
            "makefire2", ["steel", "flint"], ["fire"], and_destroy_it=False, expected_count=2
        )

        # although there is steel and flint
        # those aren't valid puzzle parts because
        # the puzzle hasn't been armed
        self._use("steel", "You have no idea how this can be used")
        self._use("steel, flint", "You have no idea how these can be used")
        self._arm(recipe_dbref, "makefire", ["steel", "flint"])
        self._check_room_contents({"steel": 2, "flint": 2}, check_test_tags=True)

        # there are duplicated objects now
        self._use("steel", "Which steel. There are many")
        self._use("flint", "Which flint. There are many")

        # delete proto parts and proto results
        self.steel.delete()
        self.flint.delete()
        self.fire.delete()

        # solve puzzle
        self._use("steel, flint", "You are a Genius")
        self.assertEqual(
            1,
            len(
                list(
                    filter(
                        lambda o: o.key == "fire"
                        and ("makefire", puzzles._PUZZLES_TAG_CATEGORY)
                        in o.tags.all(return_key_and_category=True)
                        and (puzzles._PUZZLES_TAG_MEMBER, puzzles._PUZZLES_TAG_CATEGORY)
                        in o.tags.all(return_key_and_category=True),
                        self.room1.contents,
                    )
                )
            ),
        )
        self._check_room_contents({"steel": 0, "flint": 0, "fire": 1}, check_test_tags=True)

        # trying again will fail as it was resolved already
        # and the parts were destroyed
        self._use("steel, flint", "There is no steel around")
        self._use("flint, steel", "There is no flint around")

        # arm same puzzle twice so there are duplicated parts
        self._arm(recipe_dbref, "makefire", ["steel", "flint"])
        self._arm(recipe_dbref, "makefire", ["steel", "flint"])
        self._check_room_contents({"steel": 2, "flint": 2, "fire": 1}, check_test_tags=True)

        # try solving with multiple parts but incomplete set
        self._use(
            "steel-1, steel-2", "You try to utilize these but nothing happens ... something amiss?"
        )

        # arm the other puzzle. Their parts are identical
        self._arm(recipe2_dbref, "makefire2", ["steel", "flint"])
        self._check_room_contents({"steel": 3, "flint": 3, "fire": 1}, check_test_tags=True)

        # solve with multiple parts for
        # multiple puzzles. Both can be solved but
        # only one is.
        self._use(
            "steel-1, flint-2, steel-3, flint-3",
            "Your gears start turning and 2 different ideas come to your mind ... ",
        )
        self._check_room_contents({"steel": 2, "flint": 2, "fire": 2}, check_test_tags=True)

        self.room1.msg_contents = Mock()

        # solve all
        self._use("steel-1, flint-1", "You are a Genius")
        self.room1.msg_contents.assert_called_once_with(
            "|cChar|n performs some kind of tribal dance and |yfire|n seems to appear from thin air",
            exclude=(self.char1,),
        )
        self._use("steel, flint", "You are a Genius")
        self._check_room_contents({"steel": 0, "flint": 0, "fire": 4}, check_test_tags=True)

    def test_puzzleedit(self):
        recipe_dbref = self._good_recipe(
            "makefire", ["steel", "flint"], ["fire"], and_destroy_it=False
        )

        def _puzzleedit(swt, dbref, args, expmsg):
            if (swt is None) and (dbref is None) and (args is None):
                cmdstr = ""
            else:
                cmdstr = "%s %s%s" % (swt, dbref, args)
            self.call(puzzles.CmdEditPuzzle(), cmdstr, expmsg, caller=self.char1)

        # delete proto parts and proto results
        self.steel.delete()
        self.flint.delete()
        self.fire.delete()

        sid = self.script.id
        # bad syntax
        _puzzleedit(
            None, None, None, "A puzzle recipe's #dbref must be specified.\nUsage: @puzzleedit"
        )
        _puzzleedit("", "1", "", "A puzzle recipe's #dbref must be specified.\nUsage: @puzzleedit")
        _puzzleedit("", "", "", "A puzzle recipe's #dbref must be specified.\nUsage: @puzzleedit")
        _puzzleedit(
            "",
            recipe_dbref,
            "dummy",
            "A puzzle recipe's #dbref must be specified.\nUsage: @puzzleedit",
        )
        _puzzleedit("", self.script.dbref, "", "Script(#{}) is not a puzzle".format(sid))

        # edit use_success_message and use_success_location_message
        _puzzleedit(
            "",
            recipe_dbref,
            "/use_success_message = Yes!",
            "makefire(%s) use_success_message = Yes!" % recipe_dbref,
        )
        _puzzleedit(
            "",
            recipe_dbref,
            "/use_success_location_message = {result_names} Yeah baby! {caller}",
            "makefire(%s) use_success_location_message = {result_names} Yeah baby! {caller}"
            % recipe_dbref,
        )

        self._arm(recipe_dbref, "makefire", ["steel", "flint"])
        self.room1.msg_contents = Mock()
        self._use("steel, flint", "Yes!")
        self.room1.msg_contents.assert_called_once_with(
            "fire Yeah baby! Char", exclude=(self.char1,)
        )
        self.room1.msg_contents.reset_mock()

        # edit mask: exclude location and desc during matching
        _puzzleedit(
            "",
            recipe_dbref,
            "/mask = location,desc",
            "makefire(%s) mask = ('location', 'desc')" % recipe_dbref,
        )

        self._arm(recipe_dbref, "makefire", ["steel", "flint"])
        # change location and desc
        self.char1.search("steel").db.desc = "A solid bar of steel"
        self.char1.search("steel").location = self.char1
        self.char1.search("flint").db.desc = "A flint steel"
        self.char1.search("flint").location = self.char1
        self._use("steel, flint", "Yes!")
        self.room1.msg_contents.assert_called_once_with(
            "fire Yeah baby! Char", exclude=(self.char1,)
        )

        # delete
        _puzzleedit("/delete", recipe_dbref, "", "makefire(%s) was deleted" % recipe_dbref)
        self._assert_no_recipes()

    def test_puzzleedit_add_remove_parts_results(self):
        recipe_dbref = self._good_recipe(
            "makefire", ["steel", "flint"], ["fire"], and_destroy_it=False
        )

        def _puzzleedit(swt, dbref, rhslist, expmsg):
            cmdstr = "%s %s = %s" % (swt, dbref, ", ".join(rhslist))
            self.call(puzzles.CmdEditPuzzle(), cmdstr, expmsg, caller=self.char1)

        red_steel = create_object(
            self.object_typeclass, key="red steel", location=self.char1.location
        )
        smoke = create_object(self.object_typeclass, key="smoke", location=self.char1.location)

        _puzzleedit("/addresult", recipe_dbref, ["smoke"], "smoke were added to results")
        _puzzleedit(
            "/addpart", recipe_dbref, ["red steel", "steel"], "red steel, steel were added to parts"
        )

        # create a box so we can put all objects in
        # so that they can't be found during puzzle resolution
        self.box = create_object(self.object_typeclass, key="box", location=self.char1.location)

        def _box_all():
            for o in self.room1.contents:
                if o not in [self.char1, self.char2, self.exit, self.obj1, self.obj2, self.box]:
                    o.location = self.box

        _box_all()

        self._arm(recipe_dbref, "makefire", ["steel", "flint", "red steel", "steel"])
        self._check_room_contents({"steel": 2, "red steel": 1, "flint": 1})
        self._use(
            "steel-1, flint", "You try to utilize these but nothing happens ... something amiss?"
        )
        self._use("steel-1, flint, red steel, steel-2", "You are a Genius")
        self._check_room_contents({"smoke": 1, "fire": 1})
        _box_all()

        self.fire.location = self.room1
        self.steel.location = self.room1

        _puzzleedit("/delresult", recipe_dbref, ["fire"], "fire were removed from results")
        _puzzleedit(
            "/delpart", recipe_dbref, ["steel", "steel"], "steel, steel were removed from parts"
        )

        _box_all()

        self._arm(recipe_dbref, "makefire", ["flint", "red steel"])
        self._check_room_contents({"red steel": 1, "flint": 1})
        self._use("red steel, flint", "You are a Genius")
        self._check_room_contents({"smoke": 1, "fire": 0})

    def test_lspuzzlerecipes_lsarmedpuzzles(self):
        msg = self.call(puzzles.CmdListPuzzleRecipes(), "", caller=self.char1)
        self._assert_msg_matched(
            msg, [r"^-+$", r"^Found 0 puzzle\(s\)\.$", r"-+$"], re.MULTILINE | re.DOTALL
        )

        recipe_dbref = self._good_recipe(
            "makefire", ["steel", "flint"], ["fire"], and_destroy_it=False
        )

        msg = self.call(puzzles.CmdListPuzzleRecipes(), "", caller=self.char1)
        self._assert_msg_matched(
            msg,
            [
                r"^-+$",
                r"^Puzzle 'makefire'.*$",
                r"^Success Caller message:$",
                r"^Success Location message:$",
                r"^Mask:$",
                r"^Parts$",
                r"^.*key: steel$",
                r"^.*key: flint$",
                r"^Results$",
                r"^.*key: fire$",
                r"^.*key: steel$",
                r"^.*key: flint$",
                r"^-+$",
                r"^Found 1 puzzle\(s\)\.$",
                r"^-+$",
            ],
            re.MULTILINE | re.DOTALL,
        )

        msg = self.call(puzzles.CmdListArmedPuzzles(), "", caller=self.char1)
        self._assert_msg_matched(
            msg,
            [r"^-+$", r"^-+$", r"^Found 0 armed puzzle\(s\)\.$", r"^-+$"],
            re.MULTILINE | re.DOTALL,
        )

        self._arm(recipe_dbref, "makefire", ["steel", "flint"])

        msg = self.call(puzzles.CmdListArmedPuzzles(), "", caller=self.char1)
        self._assert_msg_matched(
            msg,
            [
                r"^-+$",
                r"^Puzzle name: makefire$",
                r"^.*steel.* at \s+ Room.*$",
                r"^.*flint.* at \s+ Room.*$",
                r"^Found 1 armed puzzle\(s\)\.$",
                r"^-+$",
            ],
            re.MULTILINE | re.DOTALL,
        )

    def test_e2e(self):
        def _destroy_objs_in_room(keys):
            for obj in self.room1.contents:
                if obj.key in keys:
                    obj.delete()

        # parts don't survive resolution
        # but produce a large result set
        tree = create_object(self.object_typeclass, key="tree", location=self.char1.location)
        axe = create_object(self.object_typeclass, key="axe", location=self.char1.location)
        sweat = create_object(self.object_typeclass, key="sweat", location=self.char1.location)
        dull_axe = create_object(
            self.object_typeclass, key="dull axe", location=self.char1.location
        )
        timber = create_object(self.object_typeclass, key="timber", location=self.char1.location)
        log = create_object(self.object_typeclass, key="log", location=self.char1.location)
        parts = ["tree", "axe"]
        results = (["sweat"] * 10) + ["dull axe"] + (["timber"] * 20) + (["log"] * 50)
        recipe_dbref = self._good_recipe("lumberjack", parts, results, and_destroy_it=False)

        _destroy_objs_in_room(set(parts + results))

        sps = sorted(parts)
        expected = {key: len(list(grp)) for key, grp in itertools.groupby(sps)}
        expected.update({r: 0 for r in set(results)})

        self._arm(recipe_dbref, "lumberjack", parts)
        self._check_room_contents(expected)

        self._use(",".join(parts), "You are a Genius")
        srs = sorted(set(results))
        expected = {(key, len(list(grp))) for key, grp in itertools.groupby(srs)}
        expected.update({p: 0 for p in set(parts)})
        self._check_room_contents(expected)

        # parts also appear in results
        # causing a new puzzle to be armed 'automatically'
        # i.e. the puzzle is self-sustaining
        hole = create_object(self.object_typeclass, key="hole", location=self.char1.location)
        shovel = create_object(self.object_typeclass, key="shovel", location=self.char1.location)
        dirt = create_object(self.object_typeclass, key="dirt", location=self.char1.location)

        parts = ["shovel", "hole"]
        results = ["dirt", "hole", "shovel"]
        recipe_dbref = self._good_recipe(
            "digger", parts, results, and_destroy_it=False, expected_count=2
        )

        _destroy_objs_in_room(set(parts + results))

        nresolutions = 0

        sps = sorted(set(parts))
        expected = {key: len(list(grp)) for key, grp in itertools.groupby(sps)}
        expected.update({"dirt": nresolutions})

        self._arm(recipe_dbref, "digger", parts)
        self._check_room_contents(expected)

        for i in range(10):
            self._use(",".join(parts), "You are a Genius")
            nresolutions += 1
            expected.update({"dirt": nresolutions})
            self._check_room_contents(expected)

        # Uppercase puzzle name
        balloon = create_object(self.object_typeclass, key="Balloon", location=self.char1.location)
        parts = ["Balloon"]
        results = ["Balloon"]
        recipe_dbref = self._good_recipe(
            "boom!!!", parts, results, and_destroy_it=False, expected_count=3
        )

        _destroy_objs_in_room(set(parts + results))

        sps = sorted(parts)
        expected = {key: len(list(grp)) for key, grp in itertools.groupby(sps)}

        self._arm(recipe_dbref, "boom!!!", parts)
        self._check_room_contents(expected)

        self._use(",".join(parts), "You are a Genius")
        srs = sorted(set(results))
        expected = {(key, len(list(grp))) for key, grp in itertools.groupby(srs)}
        self._check_room_contents(expected)

    def test_e2e_accumulative(self):
        flashlight = create_object(
            self.object_typeclass, key="flashlight", location=self.char1.location
        )
        flashlight_w_1 = create_object(
            self.object_typeclass, key="flashlight-w-1", location=self.char1.location
        )
        flashlight_w_2 = create_object(
            self.object_typeclass, key="flashlight-w-2", location=self.char1.location
        )
        flashlight_w_3 = create_object(
            self.object_typeclass, key="flashlight-w-3", location=self.char1.location
        )
        battery = create_object(self.object_typeclass, key="battery", location=self.char1.location)

        battery.tags.add("flashlight-1", category=puzzles._PUZZLES_TAG_CATEGORY)
        battery.tags.add("flashlight-2", category=puzzles._PUZZLES_TAG_CATEGORY)
        battery.tags.add("flashlight-3", category=puzzles._PUZZLES_TAG_CATEGORY)

        # TODO: instead of tagging each flashlight,
        # arm and resolve each puzzle in order so they all
        # are tagged correctly
        # it will be necessary to add/remove parts/results because
        # each battery is supposed to be consumed during resolution
        # as the new flashlight has one more battery than before
        flashlight_w_1.tags.add("flashlight-2", category=puzzles._PUZZLES_TAG_CATEGORY)
        flashlight_w_2.tags.add("flashlight-3", category=puzzles._PUZZLES_TAG_CATEGORY)

        recipe_fl1_dbref = self._good_recipe(
            "flashlight-1",
            ["flashlight", "battery"],
            ["flashlight-w-1"],
            and_destroy_it=False,
            expected_count=1,
        )
        recipe_fl2_dbref = self._good_recipe(
            "flashlight-2",
            ["flashlight-w-1", "battery"],
            ["flashlight-w-2"],
            and_destroy_it=False,
            expected_count=2,
        )
        recipe_fl3_dbref = self._good_recipe(
            "flashlight-3",
            ["flashlight-w-2", "battery"],
            ["flashlight-w-3"],
            and_destroy_it=False,
            expected_count=3,
        )

        # delete protoparts
        for obj in [battery, flashlight, flashlight_w_1, flashlight_w_2, flashlight_w_3]:
            obj.delete()

        def _group_parts(parts, excluding=set()):
            group = dict()
            dbrefs = dict()
            for o in self.room1.contents:
                if o.key in parts and o.dbref not in excluding:
                    if o.key not in group:
                        group[o.key] = []
                    group[o.key].append(o.dbref)
                    dbrefs[o.dbref] = o
            return group, dbrefs

        # arm each puzzle and group its parts
        self._arm(recipe_fl1_dbref, "flashlight-1", ["battery", "flashlight"])
        fl1_parts, fl1_dbrefs = _group_parts(["battery", "flashlight"])
        self._arm(recipe_fl2_dbref, "flashlight-2", ["battery", "flashlight-w-1"])
        fl2_parts, fl2_dbrefs = _group_parts(
            ["battery", "flashlight-w-1"], excluding=list(fl1_dbrefs.keys())
        )
        self._arm(recipe_fl3_dbref, "flashlight-3", ["battery", "flashlight-w-2"])
        fl3_parts, fl3_dbrefs = _group_parts(
            ["battery", "flashlight-w-2"],
            excluding=set(list(fl1_dbrefs.keys()) + list(fl2_dbrefs.keys())),
        )

        self._check_room_contents(
            {
                "battery": 3,
                "flashlight": 1,
                "flashlight-w-1": 1,
                "flashlight-w-2": 1,
                "flashlight-w-3": 0,
            }
        )

        # all batteries have identical protodefs
        battery_1 = fl1_dbrefs[fl1_parts["battery"][0]]
        battery_2 = fl2_dbrefs[fl2_parts["battery"][0]]
        battery_3 = fl3_dbrefs[fl3_parts["battery"][0]]
        protodef_battery_1 = puzzles.proto_def(battery_1, with_tags=False)
        del protodef_battery_1["prototype_key"]
        protodef_battery_2 = puzzles.proto_def(battery_2, with_tags=False)
        del protodef_battery_2["prototype_key"]
        protodef_battery_3 = puzzles.proto_def(battery_3, with_tags=False)
        del protodef_battery_3["prototype_key"]
        assert protodef_battery_1 == protodef_battery_2 == protodef_battery_3

        # each battery can be used in every other puzzle

        b1_parts_dict, b1_puzzlenames, b1_protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
            [battery_1]
        )
        _puzzles = puzzles._puzzles_by_names(b1_puzzlenames.keys())
        assert set(["flashlight-1", "flashlight-2", "flashlight-3"]) == set(
            [p.db.puzzle_name for p in _puzzles]
        )
        matched_puzzles = puzzles._matching_puzzles(_puzzles, b1_puzzlenames, b1_protodefs)
        assert 0 == len(matched_puzzles)

        b2_parts_dict, b2_puzzlenames, b2_protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
            [battery_2]
        )
        _puzzles = puzzles._puzzles_by_names(b2_puzzlenames.keys())
        assert set(["flashlight-1", "flashlight-2", "flashlight-3"]) == set(
            [p.db.puzzle_name for p in _puzzles]
        )
        matched_puzzles = puzzles._matching_puzzles(_puzzles, b2_puzzlenames, b2_protodefs)
        assert 0 == len(matched_puzzles)
        b3_parts_dict, b3_puzzlenames, b3_protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
            [battery_3]
        )
        _puzzles = puzzles._puzzles_by_names(b3_puzzlenames.keys())
        assert set(["flashlight-1", "flashlight-2", "flashlight-3"]) == set(
            [p.db.puzzle_name for p in _puzzles]
        )
        matched_puzzles = puzzles._matching_puzzles(_puzzles, b3_puzzlenames, b3_protodefs)
        assert 0 == len(matched_puzzles)

        assert battery_1 == list(b1_parts_dict.values())[0]
        assert battery_2 == list(b2_parts_dict.values())[0]
        assert battery_3 == list(b3_parts_dict.values())[0]
        assert b1_puzzlenames.keys() == b2_puzzlenames.keys() == b3_puzzlenames.keys()
        for puzzle_name in ["flashlight-1", "flashlight-2", "flashlight-3"]:
            assert puzzle_name in b1_puzzlenames
            assert puzzle_name in b2_puzzlenames
            assert puzzle_name in b3_puzzlenames
        assert (
            list(b1_protodefs.values())[0]
            == list(b2_protodefs.values())[0]
            == list(b3_protodefs.values())[0]
            == protodef_battery_1
            == protodef_battery_2
            == protodef_battery_3
        )

        # all flashlights have similar protodefs except their key
        flashlight_1 = fl1_dbrefs[fl1_parts["flashlight"][0]]
        flashlight_2 = fl2_dbrefs[fl2_parts["flashlight-w-1"][0]]
        flashlight_3 = fl3_dbrefs[fl3_parts["flashlight-w-2"][0]]
        protodef_flashlight_1 = puzzles.proto_def(flashlight_1, with_tags=False)
        del protodef_flashlight_1["prototype_key"]
        assert protodef_flashlight_1["key"] == "flashlight"
        del protodef_flashlight_1["key"]
        protodef_flashlight_2 = puzzles.proto_def(flashlight_2, with_tags=False)
        del protodef_flashlight_2["prototype_key"]
        assert protodef_flashlight_2["key"] == "flashlight-w-1"
        del protodef_flashlight_2["key"]
        protodef_flashlight_3 = puzzles.proto_def(flashlight_3, with_tags=False)
        del protodef_flashlight_3["prototype_key"]
        assert protodef_flashlight_3["key"] == "flashlight-w-2"
        del protodef_flashlight_3["key"]
        assert protodef_flashlight_1 == protodef_flashlight_2 == protodef_flashlight_3

        # each flashlight can only be used in its own puzzle

        f1_parts_dict, f1_puzzlenames, f1_protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
            [flashlight_1]
        )
        _puzzles = puzzles._puzzles_by_names(f1_puzzlenames.keys())
        assert set(["flashlight-1"]) == set([p.db.puzzle_name for p in _puzzles])
        matched_puzzles = puzzles._matching_puzzles(_puzzles, f1_puzzlenames, f1_protodefs)
        assert 0 == len(matched_puzzles)
        f2_parts_dict, f2_puzzlenames, f2_protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
            [flashlight_2]
        )
        _puzzles = puzzles._puzzles_by_names(f2_puzzlenames.keys())
        assert set(["flashlight-2"]) == set([p.db.puzzle_name for p in _puzzles])
        matched_puzzles = puzzles._matching_puzzles(_puzzles, f2_puzzlenames, f2_protodefs)
        assert 0 == len(matched_puzzles)
        f3_parts_dict, f3_puzzlenames, f3_protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
            [flashlight_3]
        )
        _puzzles = puzzles._puzzles_by_names(f3_puzzlenames.keys())
        assert set(["flashlight-3"]) == set([p.db.puzzle_name for p in _puzzles])
        matched_puzzles = puzzles._matching_puzzles(_puzzles, f3_puzzlenames, f3_protodefs)
        assert 0 == len(matched_puzzles)

        assert flashlight_1 == list(f1_parts_dict.values())[0]
        assert flashlight_2 == list(f2_parts_dict.values())[0]
        assert flashlight_3 == list(f3_parts_dict.values())[0]
        for puzzle_name in set(
            list(f1_puzzlenames.keys()) + list(f2_puzzlenames.keys()) + list(f3_puzzlenames.keys())
        ):
            assert puzzle_name in ["flashlight-1", "flashlight-2", "flashlight-3", "puzzle_member"]
        protodef_flashlight_1["key"] = "flashlight"
        assert list(f1_protodefs.values())[0] == protodef_flashlight_1
        protodef_flashlight_2["key"] = "flashlight-w-1"
        assert list(f2_protodefs.values())[0] == protodef_flashlight_2
        protodef_flashlight_3["key"] = "flashlight-w-2"
        assert list(f3_protodefs.values())[0] == protodef_flashlight_3

        # each battery can be matched with every other flashlight
        # to potentially resolve each puzzle
        for batt in [battery_1, battery_2, battery_3]:
            parts_dict, puzzlenames, protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
                [batt, flashlight_1]
            )
            assert set([batt.dbref, flashlight_1.dbref]) == set(puzzlenames["flashlight-1"])
            assert set([batt.dbref]) == set(puzzlenames["flashlight-2"])
            assert set([batt.dbref]) == set(puzzlenames["flashlight-3"])
            _puzzles = puzzles._puzzles_by_names(puzzlenames.keys())
            assert set(["flashlight-1", "flashlight-2", "flashlight-3"]) == set(
                [p.db.puzzle_name for p in _puzzles]
            )
            matched_puzzles = puzzles._matching_puzzles(_puzzles, puzzlenames, protodefs)
            assert 1 == len(matched_puzzles)
            parts_dict, puzzlenames, protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
                [batt, flashlight_2]
            )
            assert set([batt.dbref]) == set(puzzlenames["flashlight-1"])
            assert set([batt.dbref, flashlight_2.dbref]) == set(puzzlenames["flashlight-2"])
            assert set([batt.dbref]) == set(puzzlenames["flashlight-3"])
            _puzzles = puzzles._puzzles_by_names(puzzlenames.keys())
            assert set(["flashlight-1", "flashlight-2", "flashlight-3"]) == set(
                [p.db.puzzle_name for p in _puzzles]
            )
            matched_puzzles = puzzles._matching_puzzles(_puzzles, puzzlenames, protodefs)
            assert 1 == len(matched_puzzles)
            parts_dict, puzzlenames, protodefs = puzzles._lookups_parts_puzzlenames_protodefs(
                [batt, flashlight_3]
            )
            assert set([batt.dbref]) == set(puzzlenames["flashlight-1"])
            assert set([batt.dbref]) == set(puzzlenames["flashlight-2"])
            assert set([batt.dbref, flashlight_3.dbref]) == set(puzzlenames["flashlight-3"])
            _puzzles = puzzles._puzzles_by_names(puzzlenames.keys())
            assert set(["flashlight-1", "flashlight-2", "flashlight-3"]) == set(
                [p.db.puzzle_name for p in _puzzles]
            )
            matched_puzzles = puzzles._matching_puzzles(_puzzles, puzzlenames, protodefs)
            assert 1 == len(matched_puzzles)

        # delete all parts
        for part in (
            list(fl1_dbrefs.values()) + list(fl2_dbrefs.values()) + list(fl3_dbrefs.values())
        ):
            part.delete()

        self._check_room_contents(
            {
                "battery": 0,
                "flashlight": 0,
                "flashlight-w-1": 0,
                "flashlight-w-2": 0,
                "flashlight-w-3": 0,
            }
        )

        # arm first puzzle 3 times and group its parts so we can solve
        # all puzzles with the parts from the 1st armed
        for i in range(3):
            self._arm(recipe_fl1_dbref, "flashlight-1", ["battery", "flashlight"])
        fl1_parts, fl1_dbrefs = _group_parts(["battery", "flashlight"])

        # delete the 2 extra flashlights so we can start solving
        for flashlight_dbref in fl1_parts["flashlight"][1:]:
            fl1_dbrefs[flashlight_dbref].delete()

        self._check_room_contents(
            {
                "battery": 3,
                "flashlight": 1,
                "flashlight-w-1": 0,
                "flashlight-w-2": 0,
                "flashlight-w-3": 0,
            }
        )

        self._use("battery-1, flashlight", "You are a Genius")
        self._check_room_contents(
            {
                "battery": 2,
                "flashlight": 0,
                "flashlight-w-1": 1,
                "flashlight-w-2": 0,
                "flashlight-w-3": 0,
            }
        )

        self._use("battery-1, flashlight-w-1", "You are a Genius")
        self._check_room_contents(
            {
                "battery": 1,
                "flashlight": 0,
                "flashlight-w-1": 0,
                "flashlight-w-2": 1,
                "flashlight-w-3": 0,
            }
        )

        self._use("battery, flashlight-w-2", "You are a Genius")
        self._check_room_contents(
            {
                "battery": 0,
                "flashlight": 0,
                "flashlight-w-1": 0,
                "flashlight-w-2": 0,
                "flashlight-w-3": 1,
            }
        )

    def test_e2e_interchangeable_parts_and_results(self):
        # Parts and Results can be used in multiple puzzles
        egg = create_object(self.object_typeclass, key="egg", location=self.char1.location)
        flour = create_object(self.object_typeclass, key="flour", location=self.char1.location)
        boiling_water = create_object(
            self.object_typeclass, key="boiling water", location=self.char1.location
        )
        boiled_egg = create_object(
            self.object_typeclass, key="boiled egg", location=self.char1.location
        )
        dough = create_object(self.object_typeclass, key="dough", location=self.char1.location)
        pasta = create_object(self.object_typeclass, key="pasta", location=self.char1.location)

        # Three recipes:
        # 1. breakfast: egg + boiling water = boiled egg & boiling water
        # 2. dough: egg + flour = dough
        # 3. entree: dough + boiling water = pasta & boiling water
        # tag interchangeable parts according to their puzzles' name
        egg.tags.add("breakfast", category=puzzles._PUZZLES_TAG_CATEGORY)
        egg.tags.add("dough", category=puzzles._PUZZLES_TAG_CATEGORY)
        dough.tags.add("entree", category=puzzles._PUZZLES_TAG_CATEGORY)
        boiling_water.tags.add("breakfast", category=puzzles._PUZZLES_TAG_CATEGORY)
        boiling_water.tags.add("entree", category=puzzles._PUZZLES_TAG_CATEGORY)

        # create recipes
        recipe1_dbref = self._good_recipe(
            "breakfast",
            ["egg", "boiling water"],
            ["boiled egg", "boiling water"],
            and_destroy_it=False,
        )
        recipe2_dbref = self._good_recipe(
            "dough", ["egg", "flour"], ["dough"], and_destroy_it=False, expected_count=2
        )
        recipe3_dbref = self._good_recipe(
            "entree",
            ["dough", "boiling water"],
            ["pasta", "boiling water"],
            and_destroy_it=False,
            expected_count=3,
        )

        # delete protoparts
        for obj in [egg, flour, boiling_water, boiled_egg, dough, pasta]:
            obj.delete()

        # arm each puzzle and group its parts
        def _group_parts(parts, excluding=set()):
            group = dict()
            dbrefs = dict()
            for o in self.room1.contents:
                if o.key in parts and o.dbref not in excluding:
                    if o.key not in group:
                        group[o.key] = []
                    group[o.key].append(o.dbref)
                    dbrefs[o.dbref] = o
            return group, dbrefs

        self._arm(recipe1_dbref, "breakfast", ["egg", "boiling water"])
        breakfast_parts, breakfast_dbrefs = _group_parts(["egg", "boiling water"])
        self._arm(recipe2_dbref, "dough", ["egg", "flour"])
        dough_parts, dough_dbrefs = _group_parts(
            ["egg", "flour"], excluding=list(breakfast_dbrefs.keys())
        )
        self._arm(recipe3_dbref, "entree", ["dough", "boiling water"])
        entree_parts, entree_dbrefs = _group_parts(
            ["dough", "boiling water"],
            excluding=set(list(breakfast_dbrefs.keys()) + list(dough_dbrefs.keys())),
        )

        # create a box so we can put all objects in
        # so that they can't be found during puzzle resolution
        self.box = create_object(self.object_typeclass, key="box", location=self.char1.location)

        def _box_all():
            # print "boxing all\n", "-"*20
            for o in self.room1.contents:
                if o not in [self.char1, self.char2, self.exit, self.obj1, self.obj2, self.box]:
                    o.location = self.box
                    # print o.key, o.dbref, "boxed"
                else:
                    # print "skipped", o.key, o.dbref
                    pass

        def _unbox(dbrefs):
            # print "unboxing", dbrefs, "\n", "-"*20
            for o in self.box.contents:
                if o.dbref in dbrefs:
                    o.location = self.room1
                    # print "unboxed", o.key, o.dbref

        # solve dough puzzle using breakfast's egg
        # and dough's flour. A new dough will be created
        _box_all()
        _unbox(breakfast_parts.pop("egg") + dough_parts.pop("flour"))
        self._use("egg, flour", "You are a Genius")

        # solve entree puzzle with newly created dough
        # and breakfast's boiling water. A new
        # boiling water and pasta will be created
        _unbox(breakfast_parts.pop("boiling water"))
        self._use("boiling water, dough", "You are a Genius")

        # solve breakfast puzzle with dough's egg
        # and newly created boiling water. A new
        # boiling water and boiled egg will be created
        _unbox(dough_parts.pop("egg"))
        self._use("boiling water, egg", "You are a Genius")

        # solve entree puzzle using entree's dough
        # and newly created boiling water. A new
        # boiling water and pasta will be created
        _unbox(entree_parts.pop("dough"))
        self._use("boiling water, dough", "You are a Genius")

        self._check_room_contents({"boiling water": 1, "pasta": 2, "boiled egg": 1})



"""
This is an advanced ASCII table creator. It was inspired by Prettytable
(https://code.google.com/p/prettytable/) but shares no code and is considerably
more advanced, supporting auto-balancing of incomplete tables and ANSI colors among
other things.

Example usage:

```python
  from evennia.utils import evtable

  table = evtable.EvTable("Heading1", "Heading2",
                  table=[[1,2,3],[4,5,6],[7,8,9]], border="cells")
  table.add_column("This is long data", "This is even longer data")
  table.add_row("This is a single row")
  print table
```

Result:

::

    +----------------------+----------+---+--------------------------+
    |       Heading1       | Heading2 |   |                          |
    +~~~~~~~~~~~~~~~~~~~~~~+~~~~~~~~~~+~~~+~~~~~~~~~~~~~~~~~~~~~~~~~~+
    |           1          |     4    | 7 |     This is long data    |
    +----------------------+----------+---+--------------------------+
    |           2          |     5    | 8 | This is even longer data |
    +----------------------+----------+---+--------------------------+
    |           3          |     6    | 9 |                          |
    +----------------------+----------+---+--------------------------+
    | This is a single row |          |   |                          |
    +----------------------+----------+---+--------------------------+

As seen, the table will automatically expand with empty cells to make
the table symmetric. Tables can be restricted to a given width:

```python
  table.reformat(width=50, align="l")
```

(We could just have added these keywords to the table creation call)

This yields the following result:

::

    +-----------+------------+-----------+-----------+
    | Heading1  | Heading2   |           |           |
    +~~~~~~~~~~~+~~~~~~~~~~~~+~~~~~~~~~~~+~~~~~~~~~~~+
    | 1         | 4          | 7         | This is   |
    |           |            |           | long data |
    +-----------+------------+-----------+-----------+
    |           |            |           | This is   |
    | 2         | 5          | 8         | even      |
    |           |            |           | longer    |
    |           |            |           | data      |
    +-----------+------------+-----------+-----------+
    | 3         | 6          | 9         |           |
    +-----------+------------+-----------+-----------+
    | This is a |            |           |           |
    |  single   |            |           |           |
    | row       |            |           |           |
    +-----------+------------+-----------+-----------+


Table-columns can be individually formatted. Note that if an
individual column is set with a specific width, table auto-balancing
will not affect this column (this may lead to the full table being too
wide, so be careful mixing fixed-width columns with auto- balancing).
Here we change the width and alignment of the column at index 3
(Python starts from 0):

```python

table.reformat_column(3, width=30, align="r")
print table
```

::

    +-----------+-------+-----+-----------------------------+---------+
    | Heading1  | Headi |     |                             |         |
    |           | ng2   |     |                             |         |
    +~~~~~~~~~~~+~~~~~~~+~~~~~+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+~~~~~~~~~+
    | 1         | 4     | 7   |           This is long data | Test1   |
    +-----------+-------+-----+-----------------------------+---------+
    | 2         | 5     | 8   |    This is even longer data | Test3   |
    +-----------+-------+-----+-----------------------------+---------+
    | 3         | 6     | 9   |                             | Test4   |
    +-----------+-------+-----+-----------------------------+---------+
    | This is a |       |     |                             |         |
    |  single   |       |     |                             |         |
    | row       |       |     |                             |         |
    +-----------+-------+-----+-----------------------------+---------+

When adding new rows/columns their data can have its own alignments
(left/center/right, top/center/bottom).

If the height is restricted, cells will be restricted from expanding
vertically. This will lead to text contents being cropped. Each cell
can only shrink to a minimum width and height of 1.

`EvTable` is intended to be used with `ANSIString` for supporting ANSI-coloured
string types.

When a cell is auto-wrapped across multiple lines, ANSI-reset sequences will be
put at the end of each wrapped line. This means that the colour of a wrapped
cell will not "bleed", but it also means that eventual colour outside the table
will not transfer "across" a table, you need to re-set the color to have it
appear on both sides of the table string.

----

"""

from copy import copy, deepcopy
from textwrap import TextWrapper

from django.conf import settings

from evennia.utils.ansi import ANSIString
from evennia.utils.utils import display_len as d_len
from evennia.utils.utils import is_iter, justify

_DEFAULT_WIDTH = settings.CLIENT_DEFAULT_WIDTH


def _to_ansi(obj):
    """
    convert to ANSIString.

    Args:
        obj (str): Convert incoming text to
            be ANSI aware ANSIStrings.
    """
    if is_iter(obj):
        return [_to_ansi(o) for o in obj]
    else:
        return ANSIString(obj)


_whitespace = "\t\n\x0b\x0c\r "


class ANSITextWrapper(TextWrapper):
    """
    This is a wrapper work class for handling strings with ANSI tags
    in it.  It overloads the standard library `TextWrapper` class and
    is used internally in `EvTable` and has no public methods.

    """

    def _munge_whitespace(self, text):
        """_munge_whitespace(text : string) -> string

        Munge whitespace in text: expand tabs and convert all other
        whitespace characters to spaces.  Eg. " foo\tbar\n\nbaz"
        becomes " foo    bar  baz".
        """
        return text

    # TODO: Ignore expand_tabs/replace_whitespace until ANSIString handles them.
    # - don't remove this code. /Griatch
    #        if self.expand_tabs:
    #            text = text.expandtabs()
    #        if self.replace_whitespace:
    #            if isinstance(text, str):
    #                text = text.translate(self.whitespace_trans)
    #        return text

    def _split(self, text):
        """_split(text : string) -> [string]

        Split the text to wrap into indivisible chunks.  Chunks are
        not quite the same as words; see _wrap_chunks() for full
        details.  As an example, the text
          Look, goof-ball -- use the -b option!
        breaks into the following chunks:
          'Look,', ' ', 'goof-', 'ball', ' ', '--', ' ',
          'use', ' ', 'the', ' ', '-b', ' ', 'option!'
        if break_on_hyphens is True, or in:
          'Look,', ' ', 'goof-ball', ' ', '--', ' ',
          'use', ' ', 'the', ' ', '-b', ' ', option!'
        otherwise.
        """
        # NOTE-PYTHON3: The following code only roughly approximates what this
        #               function used to do. Regex splitting on ANSIStrings is
        #               dropping ANSI codes, so we're using ANSIString.split
        #               for the time being.
        #
        #               A less hackier solution would be appreciated.
        chunks = _to_ansi(text).split()

        chunks = [chunk + " " for chunk in chunks if chunk]  # remove empty chunks

        if len(chunks) > 1:
            chunks[-1] = chunks[-1][0:-1]

        return chunks

    def _wrap_chunks(self, chunks):
        """_wrap_chunks(chunks : [string]) -> [string]

        Wrap a sequence of text chunks and return a list of lines of
        length 'self.width' or less.  (If 'break_long_words' is false,
        some lines may be longer than this.)  Chunks correspond roughly
        to words and the whitespace between them: each chunk is
        indivisible (modulo 'break_long_words'), but a line break can
        come between any two chunks.  Chunks should not have internal
        whitespace; ie. a chunk is either all whitespace or a "word".
        Whitespace chunks will be removed from the beginning and end of
        lines, but apart from that whitespace is preserved.
        """
        lines = []
        if self.width <= 0:
            raise ValueError("invalid width %r (must be > 0)" % self.width)

        # Arrange in reverse order so items can be efficiently popped
        # from a stack of chucks.
        chunks.reverse()

        while chunks:
            # Start the list of chunks that will make up the current line.
            # cur_len is just the length of all the chunks in cur_line.
            cur_line = []
            cur_len = 0

            # Figure out which static string will prefix this line.
            if lines:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent

            # Maximum width for this line.
            width = self.width - d_len(indent)

            # First chunk on line is whitespace -- drop it, unless this
            # is the very beginning of the text (ie. no lines started yet).
            if self.drop_whitespace and chunks[-1].strip() == "" and lines:
                del chunks[-1]

            while chunks:
                ln = d_len(chunks[-1])

                # Can at least squeeze this chunk onto the current line.
                if cur_len + ln <= width:
                    cur_line.append(chunks.pop())
                    cur_len += ln

                # Nope, this line is full.
                else:
                    break

            # The current line is full, and the next chunk is too big to
            # fit on *any* line (not just this one).
            if chunks and d_len(chunks[-1]) > width:
                self._handle_long_word(chunks, cur_line, cur_len, width)

            # If the last chunk on this line is all whitespace, drop it.
            if self.drop_whitespace and cur_line and cur_line[-1].strip() == "":
                del cur_line[-1]

            # Convert current line back to a string and store it in list
            # of all lines (return value).
            if cur_line:
                ln = ""
                for w in cur_line:  # ANSI fix
                    ln += w  #
                lines.append(indent + ln)
        return lines


# -- Convenience interface ---------------------------------------------


def wrap(text, width=_DEFAULT_WIDTH, **kwargs):
    """
    Wrap a single paragraph of text, returning a list of wrapped lines.

    Reformat the single paragraph in 'text' so it fits in lines of no
    more than 'width' columns, and return a list of wrapped lines.  By
    default, tabs in 'text' are expanded with string.expandtabs(), and
    all other whitespace characters (including newline) are converted to

    Args:
        text (str): Text to wrap.
        width (int, optional): Width to wrap `text` to.

    Keyword Args:
        See TextWrapper class for available keyword args to customize
        wrapping behaviour.

    """
    w = ANSITextWrapper(width=width, **kwargs)
    return w.wrap(text)


def fill(text, width=_DEFAULT_WIDTH, **kwargs):
    """Fill a single paragraph of text, returning a new string.

    Reformat the single paragraph in 'text' to fit in lines of no more
    than 'width' columns, and return a new string containing the entire
    wrapped paragraph.  As with wrap(), tabs are expanded and other
    whitespace characters converted to space.

    Args:
        text (str): Text to fill.
        width (int, optional): Width of fill area.

    Keyword Args:
        See TextWrapper class for available keyword args to customize
        filling behaviour.

    """
    w = ANSITextWrapper(width=width, **kwargs)
    return w.fill(text)


# EvCell class (see further down for the EvTable itself)


class EvCell:
    """
    Holds a single data cell for the table. A cell has a certain width
    and height and contains one or more lines of data. It can shrink
    and resize as needed.

    """

    def __init__(self, data, **kwargs):
        """
        Args:
            data (str): The un-padded data of the entry.

        Keyword Args:
            width (int): Desired width of cell. It will pad
                to this size.
            height (int): Desired height of cell. it will pad
                to this size.
            pad_width (int): General padding width. This can be overruled
                by individual settings below.
            pad_left (int): Number of extra pad characters on the left.
            pad_right (int): Number of extra pad characters on the right.
            pad_top (int):  Number of extra pad lines top (will pad with `vpad_char`).
            pad_bottom (int): Number of extra pad lines bottom (will pad with `vpad_char`).
            pad_char (str)- pad character to use for padding. This is overruled
                by individual settings below (default `" "`).
            hpad_char (str): Pad character to use both for extra horizontal
                padding (default `" "`).
            vpad_char (str): Pad character to use for extra vertical padding
                and for vertical fill (default `" "`).
            fill_char (str): Character used to filling (expanding cells to
                desired size). This can be overruled by individual settings below.
            hfill_char (str): Character used for horizontal fill (default `" "`).
            vfill_char (str): Character used for vertical fill (default `" "`).
            align (str): Should be one of "l", "r", "c", "f" or "a" for left-, right-, center-,
                full-justified (with space between words) or absolute (keep as much original
                whitespace as possible). Default is left-aligned.
            valign (str): Should be one of "t", "b" or "c" for top-, bottom and center
                vertical alignment respectively. Default is centered.
            border_width (int): General border width. This is overruled
                by individual settings below.
            border_left (int): Left border width.
            border_right (int): Right border width.
            border_top (int): Top border width.
            border_bottom (int): Bottom border width.
            border_char (str): This will use a single border char for all borders.
                  overruled by individual settings below.
            border_left_char (str): Char used for left border.
            border_right_char (str): Char used for right border.
            border_top_char (str): Char used for top border.
            border_bottom_char (str): Char user for bottom border.
            corner_char (str): Character used when two borders cross.  (default is "").
                This is overruled by individual settings below.
            corner_top_left_char (str): Char used for "nw" corner.
            corner_top_right_char (str):  Char used for "ne" corner.
            corner_bottom_left_char (str): Char used for "sw" corner.
            corner_bottom_right_char (str): Char used for "se" corner.
            crop_string (str): String to use when cropping sideways, default is `'[...]'`.
            crop (bool): Crop contentof cell rather than expand vertically, default=`False`.
            enforce_size (bool): If true, the width/height of the cell is
                strictly enforced and extra text will be cropped rather than the
                cell growing vertically.

        Raises:
            Exception: for impossible cell size requirements where the
                border width or height cannot fit, or the content is too
                small.

        """
        self.formatted = None
        padwidth = kwargs.get("pad_width", None)
        padwidth = int(padwidth) if padwidth is not None else None
        self.pad_left = int(kwargs.get("pad_left", padwidth if padwidth is not None else 1))
        self.pad_right = int(kwargs.get("pad_right", padwidth if padwidth is not None else 1))
        self.pad_top = int(kwargs.get("pad_top", padwidth if padwidth is not None else 0))
        self.pad_bottom = int(kwargs.get("pad_bottom", padwidth if padwidth is not None else 0))

        self.enforce_size = kwargs.get("enforce_size", False)

        # avoid multi-char pad_chars messing up counting
        pad_char = kwargs.get("pad_char", " ")
        pad_char = pad_char[0] if pad_char else " "
        hpad_char = kwargs.get("hpad_char", pad_char)
        self.hpad_char = hpad_char[0] if hpad_char else pad_char
        vpad_char = kwargs.get("vpad_char", pad_char)
        self.vpad_char = vpad_char[0] if vpad_char else pad_char

        fill_char = kwargs.get("fill_char", " ")
        fill_char = fill_char[0] if fill_char else " "
        hfill_char = kwargs.get("hfill_char", fill_char)
        self.hfill_char = hfill_char[0] if hfill_char else " "
        vfill_char = kwargs.get("vfill_char", fill_char)
        self.vfill_char = vfill_char[0] if vfill_char else " "

        self.crop_string = kwargs.get("crop_string", "[...]")

        # borders and corners
        borderwidth = kwargs.get("border_width", 0)
        self.border_left = kwargs.get("border_left", borderwidth)
        self.border_right = kwargs.get("border_right", borderwidth)
        self.border_top = kwargs.get("border_top", borderwidth)
        self.border_bottom = kwargs.get("border_bottom", borderwidth)

        borderchar = kwargs.get("border_char", None)
        self.border_left_char = kwargs.get("border_left_char", borderchar if borderchar else "|")
        self.border_right_char = kwargs.get(
            "border_right_char", borderchar if borderchar else self.border_left_char
        )
        self.border_top_char = kwargs.get("border_top_char", borderchar if borderchar else "-")
        self.border_bottom_char = kwargs.get(
            "border_bottom_char", borderchar if borderchar else self.border_top_char
        )

        corner_char = kwargs.get("corner_char", "+")
        self.corner_top_left_char = kwargs.get("corner_top_left_char", corner_char)
        self.corner_top_right_char = kwargs.get("corner_top_right_char", corner_char)
        self.corner_bottom_left_char = kwargs.get("corner_bottom_left_char", corner_char)
        self.corner_bottom_right_char = kwargs.get("corner_bottom_right_char", corner_char)

        # alignments
        self.align = kwargs.get("align", "l")
        self.valign = kwargs.get("valign", "c")

        self.data = self._split_lines(_to_ansi(data))
        self.raw_width = max(d_len(line) for line in self.data)
        self.raw_height = len(self.data)

        # this is extra trimming required for cels in the middle of a table only
        self.trim_horizontal = 0
        self.trim_vertical = 0

        # width/height is given without left/right or top/bottom padding
        if "width" in kwargs:
            width = kwargs.pop("width")
            self.width = (
                width - self.pad_left - self.pad_right - self.border_left - self.border_right
            )
            if self.width <= 0 < self.raw_width:
                raise Exception("Cell width too small - no space for data.")
        else:
            self.width = self.raw_width
        if "height" in kwargs:
            height = kwargs.pop("height")
            self.height = (
                height - self.pad_top - self.pad_bottom - self.border_top - self.border_bottom
            )
            if self.height <= 0 < self.raw_height:
                raise Exception("Cell height too small - no space for data.")
        else:
            self.height = self.raw_height

    def _reformat(self):
        """
        Apply all EvCells' formatting operations.

        """
        data = self._border(self._pad(self._valign(self._align(self._fit_width(self.data)))))
        return data

    def _split_lines(self, text):
        """
        Simply split by linebreaks

        Args:
            text (str): text to split.

        Returns:
            split (list): split text.
        """
        return text.split("\n")

    def _fit_width(self, data):
        """
        Split too-long lines to fit the desired width of the Cell.

        Args:
            data (str): Text to adjust to the cell's width.

        Returns:
            adjusted data (str): The adjusted text.

        Notes:
            This also updates `raw_width`.


        """
        width = self.width
        adjusted_data = []
        for line in data:
            if 0 < width < d_len(line):
                # replace_whitespace=False, expand_tabs=False is a
                # fix for ANSIString not supporting expand_tabs/translate
                adjusted_data.extend(
                    [
                        ANSIString(part + ANSIString("|n"))
                        for part in wrap(line, width=width, drop_whitespace=False)
                    ]
                )
            else:
                adjusted_data.append(line)
        if self.enforce_size:
            # don't allow too high cells
            excess = len(adjusted_data) - self.height
            if excess > 0:
                # too many lines. Crop and mark last line with crop_string
                crop_string = self.crop_string
                adjusted_data = adjusted_data[:-excess]
                adjusted_data_length = len(adjusted_data[-1])
                crop_string_length = len(crop_string)
                if adjusted_data_length >= crop_string_length:
                    # replace with data[...]
                    # (note that if adjusted data is shorter than the crop-string,
                    # we skip the crop-string and just pass the cropped data.)
                    adjusted_data[-1] = adjusted_data[-1][:-crop_string_length] + crop_string

            elif excess < 0:
                # too few lines. Fill to height.
                adjusted_data.extend(["" for _ in range(excess)])

        return adjusted_data

    def _align(self, data):
        """
        Align list of rows of cell. Whitespace characters will be stripped
        if there is only one whitespace character - otherwise, it's assumed
        the caller may be trying some manual formatting in the text.

        Args:
            data (str): Text to align.

        Returns:
            text (str): Aligned result.

        """
        align = self.align
        hfill_char = self.hfill_char
        width = self.width
        return [justify(line, width, align=align, fillchar=hfill_char) for line in data]

    def _valign(self, data):
        """
        Align cell vertically

        Args:
            data (str): Text to align.

        Returns:
            text (str): Vertically aligned text.

        """
        valign = self.valign
        height = self.height
        cheight = len(data)
        excess = height - cheight
        padline = self.vfill_char * self.width

        if excess <= 0:
            return data
        # only care if we need to add new lines
        if valign == "t":
            return data + [padline for _ in range(excess)]
        elif valign == "b":
            return [padline for _ in range(excess)] + data
        else:  # center
            narrowside = [padline for _ in range(excess // 2)]
            widerside = narrowside + [padline]
            if excess % 2:
                # uneven padding
                if height % 2:
                    return widerside + data + narrowside
                else:
                    return narrowside + data + widerside
            else:
                # even padding, same on both sides
                return narrowside + data + narrowside

    def _pad(self, data):
        """
        Pad data with extra characters on all sides.

        Args:
            data (str): Text to pad.

        Returns:
            text (str): Padded text.

        """
        left = self.hpad_char * self.pad_left
        right = self.hpad_char * self.pad_right
        vfill = (self.width + self.pad_left + self.pad_right) * self.vpad_char
        top = [vfill for _ in range(self.pad_top)]
        bottom = [vfill for _ in range(self.pad_bottom)]
        return top + [left + line + right for line in data] + bottom

    def _border(self, data):
        """
        Add borders to the cell.

        Args:
            data (str): Text to surround with borders.

        Return:
            text (str): Text with borders.

        """

        left = self.border_left_char * self.border_left + ANSIString("|n")
        right = ANSIString("|n") + self.border_right_char * self.border_right

        cwidth = (
            self.width
            + self.pad_left
            + self.pad_right
            + max(0, self.border_left - 1)
            + max(0, self.border_right - 1)
        )

        vfill = self.corner_top_left_char if left else ""
        vfill += cwidth * self.border_top_char
        vfill += self.corner_top_right_char if right else ""
        top = [vfill for _ in range(self.border_top)]

        vfill = self.corner_bottom_left_char if left else ""
        vfill += cwidth * self.border_bottom_char
        vfill += self.corner_bottom_right_char if right else ""
        bottom = [vfill for _ in range(self.border_bottom)]

        return top + [left + line + right for line in data] + bottom

    def get_min_height(self):
        """
        Get the minimum possible height of cell, including at least
        one line for data.

        Returns:
            min_height (int): The mininum height of cell.

        """
        return self.pad_top + self.pad_bottom + self.border_bottom + self.border_top + 1

    def get_min_width(self):
        """
        Get the minimum possible width of cell, including at least one
        character-width for data.

        Returns:
            min_width (int): The minimum width of cell.

        """
        return self.pad_left + self.pad_right + self.border_left + self.border_right + 1

    def get_height(self):
        """
        Get natural height of cell, including padding.

        Returns:
            natural_height (int): Height of cell.

        """
        return len(self.formatted)  # if self.formatted else 0

    def get_width(self):
        """
        Get natural width of cell, including padding.

        Returns:
            natural_width (int): Width of cell.

        """
        return d_len(self.formatted[0])  # if self.formatted else 0

    def replace_data(self, data, **kwargs):
        """
        Replace cell data. This causes a full reformat of the cell.

        Args:
            data (str): Cell data.

        Notes:
            The available keyword arguments are the same as for
            `EvCell.__init__`.

        """
        self.data = self._split_lines(_to_ansi(data))
        self.raw_width = max(d_len(line) for line in self.data)
        self.raw_height = len(self.data)
        self.reformat(**kwargs)

    def reformat(self, **kwargs):
        """
        Reformat the EvCell with new options

        Keyword Args:
            The available keyword arguments are the same as for `EvCell.__init__`.

        Raises:
            Exception: If the cells cannot shrink enough to accomodate
                the options or the data given.

        """
        # keywords that require manipulation
        padwidth = kwargs.get("pad_width", None)
        padwidth = int(padwidth) if padwidth is not None else None
        self.pad_left = int(
            kwargs.pop("pad_left", padwidth if padwidth is not None else self.pad_left)
        )
        self.pad_right = int(
            kwargs.pop("pad_right", padwidth if padwidth is not None else self.pad_right)
        )
        self.pad_top = int(
            kwargs.pop("pad_top", padwidth if padwidth is not None else self.pad_top)
        )
        self.pad_bottom = int(
            kwargs.pop("pad_bottom", padwidth if padwidth is not None else self.pad_bottom)
        )

        self.enforce_size = kwargs.get("enforce_size", False)

        pad_char = kwargs.pop("pad_char", None)
        hpad_char = kwargs.pop("hpad_char", pad_char)
        self.hpad_char = hpad_char[0] if hpad_char else self.hpad_char
        vpad_char = kwargs.pop("vpad_char", pad_char)
        self.vpad_char = vpad_char[0] if vpad_char else self.vpad_char

        fillchar = kwargs.pop("fill_char", None)
        hfill_char = kwargs.pop("hfill_char", fillchar)
        self.hfill_char = hfill_char[0] if hfill_char else self.hfill_char
        vfill_char = kwargs.pop("vfill_char", fillchar)
        self.vfill_char = vfill_char[0] if vfill_char else self.vfill_char

        borderwidth = kwargs.get("border_width", None)
        self.border_left = kwargs.pop(
            "border_left", borderwidth if borderwidth is not None else self.border_left
        )
        self.border_right = kwargs.pop(
            "border_right", borderwidth if borderwidth is not None else self.border_right
        )
        self.border_top = kwargs.pop(
            "border_top", borderwidth if borderwidth is not None else self.border_top
        )
        self.border_bottom = kwargs.pop(
            "border_bottom", borderwidth if borderwidth is not None else self.border_bottom
        )

        borderchar = kwargs.get("border_char", None)
        self.border_left_char = kwargs.pop(
            "border_left_char", borderchar if borderchar else self.border_left_char
        )
        self.border_right_char = kwargs.pop(
            "border_right_char", borderchar if borderchar else self.border_right_char
        )
        self.border_top_char = kwargs.pop(
            "border_topchar", borderchar if borderchar else self.border_top_char
        )
        self.border_bottom_char = kwargs.pop(
            "border_bottom_char", borderchar if borderchar else self.border_bottom_char
        )

        corner_char = kwargs.get("corner_char", None)
        self.corner_top_left_char = kwargs.pop(
            "corner_top_left", corner_char if corner_char is not None else self.corner_top_left_char
        )
        self.corner_top_right_char = kwargs.pop(
            "corner_top_right",
            corner_char if corner_char is not None else self.corner_top_right_char,
        )
        self.corner_bottom_left_char = kwargs.pop(
            "corner_bottom_left",
            corner_char if corner_char is not None else self.corner_bottom_left_char,
        )
        self.corner_bottom_right_char = kwargs.pop(
            "corner_bottom_right",
            corner_char if corner_char is not None else self.corner_bottom_right_char,
        )

        # this is used by the table to adjust size of cells with borders in the middle
        # of the table
        self.trim_horizontal = kwargs.pop("trim_horizontal", self.trim_horizontal)
        self.trim_vertical = kwargs.pop("trim_vertical", self.trim_vertical)

        # fill all other properties
        for key, value in kwargs.items():
            setattr(self, key, value)

        # Handle sizes
        if "width" in kwargs:
            width = kwargs.pop("width")
            self.width = (
                width
                - self.pad_left
                - self.pad_right
                - self.border_left
                - self.border_right
                + self.trim_horizontal
            )
            # if self.width <= 0 and self.raw_width > 0:
            if self.width <= 0 < self.raw_width:
                raise Exception("Cell width too small, no room for data.")
        if "height" in kwargs:
            height = kwargs.pop("height")
            self.height = (
                height
                - self.pad_top
                - self.pad_bottom
                - self.border_top
                - self.border_bottom
                + self.trim_vertical
            )
            if self.height <= 0 < self.raw_height:
                raise Exception("Cell height too small, no room for data.")

        # reformat (to new sizes, padding, header and borders)
        self.formatted = self._reformat()

    def get(self):
        """
        Get data, padded and aligned in the form of a list of lines.

        """
        if not self.formatted:
            self.formatted = self._reformat()
        return self.formatted

    def __repr__(self):
        if not self.formatted:
            self.formatted = self._reformat()
        return str(ANSIString("<EvCel %s>" % self.formatted))

    def __str__(self):
        "returns cell contents on string form"
        if not self.formatted:
            self.formatted = self._reformat()
        return str(ANSIString("\n").join(self.formatted))


# EvColumn class


class EvColumn:
    """
    This class holds a list of Cells to represent a column of a table.
    It holds operations and settings that affect *all* cells in the
    column.

    Columns are not intended to be used stand-alone; they should be
    incorporated into an EvTable (like EvCells)

    """

    def __init__(self, *args, **kwargs):
        """
        Args:
            Text for each row in the column

        Keyword Args:
            All `EvCell.__init_` keywords are available, these
            settings will be persistently applied to every Cell in the
            column.

        """
        self.options = kwargs  # column-specific options
        self.column = [EvCell(data, **kwargs) for data in args]

    def _balance(self, **kwargs):
        """
        Make sure to adjust the width of all cells so we form a
        coherent and lined-up column. Will enforce column-specific
        options to cells.

        Keyword Args:
            Extra keywords to modify the column setting. Same keywords
            as in `EvCell.__init__`.

        """
        col = self.column
        # fixed options for the column will override those requested in the call!
        # this is particularly relevant to things like width/height, to avoid
        # fixed-widths columns from being auto-balanced
        kwargs.update(self.options)
        # use fixed width or adjust to the largest cell
        if "width" not in kwargs:
            [
                cell.reformat() for cell in col
            ]  # this is necessary to get initial widths of all cells
            kwargs["width"] = max(cell.get_width() for cell in col) if col else 0
        [cell.reformat(**kwargs) for cell in col]

    def add_rows(self, *args, **kwargs):
        """
        Add new cells to column. They will be inserted as
        a series of rows. It will inherit the options
        of the rest of the column's cells (use update to change
        options).

        Args:
            Texts for the new cells
            ypos (int, optional): Index position in table before which to insert the
                new column. Uses Python indexing, so to insert at the top,
                use `ypos=0`. If not given, data will be inserted at the end
                of the column.

        Keyword Args:
            Available keywods as per `EvCell.__init__`.

        """
        # column-level options override those in kwargs
        options = {**kwargs, **self.options}

        ypos = kwargs.get("ypos", None)
        if ypos is None or ypos > len(self.column):
            # add to the end
            self.column.extend([EvCell(data, **options) for data in args])
        else:
            # insert cells before given index
            ypos = min(len(self.column) - 1, max(0, int(ypos)))
            new_cells = [EvCell(data, **options) for data in args]
            self.column = self.column[:ypos] + new_cells + self.column[ypos:]
        # self._balance(**kwargs)

    def reformat(self, **kwargs):
        """
        Change the options for the column.

        Keyword Args:
            Keywords as per `EvCell.__init__`.

        """
        self._balance(**kwargs)

    def reformat_cell(self, index, **kwargs):
        """
        reformat cell at given index, keeping column options if
        necessary.

        Args:
            index (int): Index location of the cell in the column,
                starting from 0 for the first row to Nrows-1.

        Keyword Args:
            Keywords as per `EvCell.__init__`.

        """
        # column-level options take precedence here
        kwargs.update(self.options)
        self.column[index].reformat(**kwargs)

    def __repr__(self):
        return "<EvColumn\n  %s>" % "\n  ".join([repr(cell) for cell in self.column])

    def __len__(self):
        return len(self.column)

    def __iter__(self):
        return iter(self.column)

    def __getitem__(self, index):
        return self.column[index]

    def __setitem__(self, index, value):
        self.column[index] = value

    def __delitem__(self, index):
        del self.column[index]


# Main Evtable class


class EvTable:
    """
    The table class holds a list of EvColumns, each consisting of EvCells so
    that the result is a 2D matrix.
    """

    def __init__(self, *args, **kwargs):
        """
        Args:
            Header texts for the table.

        Keyword Args:
            table (list of lists or list of `EvColumns`, optional):
                This is used to build the table in a quick way.  If not
                given, the table will start out empty and `add_` methods
                need to be used to add rows/columns.
            header (bool, optional): `True`/`False` - turn off the
                header texts (`*args`) being treated as a header (such as
                not adding extra underlining)
            pad_width (int, optional): How much empty space to pad your cells with
                (default is 1)
            border (str, optional)): The border style to use. This is one of
                    - `None` - No border drawing at all.
                    - "table" - only a border around the whole table.
                    - "tablecols" - table and column borders. (default)
                    - "header" - only border under header.
                    - "cols" - only vertical borders.
                    - "incols" - vertical borders, no outer edges.
                    - "rows" - only borders between rows.
                    - "cells" - border around all cells.
            border_width (int, optional): Width of table borders, if border is active.
                Note that widths wider than 1 may give artifacts in the corners. Default is 1.
            corner_char (str, optional): Character to use in corners when border is active.
                Default is `+`.
            corner_top_left_char (str, optional): Character used for "nw" corner of table.
                Defaults to `corner_char`.
            corner_top_right_char (str, optional): Character used for "ne" corner of table.
                Defaults to `corner_char`.
            corner_bottom_left_char (str, optional): Character used for "sw" corner of table.
                Defaults to `corner_char`.
            corner_bottom_right_char (str, optional): Character used for "se" corner of table.
                Defaults to `corner_char`.
            pretty_corners (bool, optional): Use custom characters to
                make the table corners look "rounded". Uses UTF-8
                characters. Defaults to `False` for maximum compatibility with various displays
                that may occationally have issues with UTF-8 characters.
            header_line_char (str, optional): Character to use for underlining
                the header row (default is '~'). Requires `border` to not be `None`.
            width (int, optional): Fixed width of table. If not set,
                width is set by the total width of each column.  This will
                resize individual columns in the vertical direction to fit.
            height (int, optional): Fixed height of table. Defaults to being unset. Width is
                still given precedence. If given, table cells will crop text rather
                than expand vertically.
            evenwidth (bool, optional): Used with the `width` keyword. Adjusts columns to have as
                even width as possible. This often looks best also for mixed-length tables. Default
                is `False`.
            maxwidth (int, optional):  This will set a maximum width
                of the table while allowing it to be smaller. Only if it grows wider than this
                size will it be resized by expanding horizontally (or crop `height` is given).
                This keyword has no meaning if `width` is set.

        Raises:
            Exception: If given erroneous input or width settings for the data.

        Notes:
            Beyond those table-specific keywords, the non-overlapping keywords
            of `EvCell.__init__` are also available. These will be passed down
            to every cell in the table.

        """
        # at this point table is a 2D grid - a list of columns
        # x is the column position, y the row
        table = kwargs.pop("table", [])

        # header is a list of texts. We merge it to the table's top
        header = [_to_ansi(head) for head in args]
        self.header = header != []
        if self.header:
            if table:
                excess = len(header) - len(table)
                if excess > 0:
                    # header bigger than table
                    table.extend([] for _ in range(excess))
                elif excess < 0:
                    # too short header
                    header.extend(_to_ansi(["" for _ in range(abs(excess))]))
                for ix, heading in enumerate(header):
                    table[ix].insert(0, heading)
            else:
                table = [[heading] for heading in header]
        # even though we inserted the header, we can still turn off
        # header border underling etc. We only allow this if a header
        # was actually set
        self.header = kwargs.pop("header", self.header) if self.header else False
        hchar = kwargs.pop("header_line_char", "~")
        self.header_line_char = hchar[0] if hchar else "~"

        border = kwargs.pop("border", "tablecols")
        if border is None:
            border = "none"
        if border not in (
            "none",
            "table",
            "tablecols",
            "header",
            "incols",
            "cols",
            "rows",
            "cells",
        ):
            raise Exception("Unsupported border type: '%s'" % border)
        self.border = border

        # border settings are passed into Cell as well (so kwargs.get and not pop)
        self.border_width = kwargs.get("border_width", 1)
        self.corner_char = kwargs.get("corner_char", "+")
        pcorners = kwargs.pop("pretty_corners", False)
        self.corner_top_left_char = _to_ansi(
            kwargs.pop("corner_top_left_char", "." if pcorners else self.corner_char)
        )
        self.corner_top_right_char = _to_ansi(
            kwargs.pop("corner_top_right_char", "." if pcorners else self.corner_char)
        )
        self.corner_bottom_left_char = _to_ansi(
            kwargs.pop("corner_bottom_left_char", " " if pcorners else self.corner_char)
        )
        self.corner_bottom_right_char = _to_ansi(
            kwargs.pop("corner_bottom_right_char", " " if pcorners else self.corner_char)
        )

        self.width = kwargs.pop("width", None)
        self.height = kwargs.pop("height", None)
        self.evenwidth = kwargs.pop("evenwidth", False)
        self.maxwidth = kwargs.pop("maxwidth", None)
        if self.maxwidth and self.width and self.maxwidth < self.width:
            raise Exception("table maxwidth < table width!")
        # size in cell cols/rows
        self.ncols = len(table)
        self.nrows = max(len(col) for col in table) if table else 0
        # size in characters (gets set when _balance is called)
        self.nwidth = 0
        self.nheight = 0
        # save options
        self.options = kwargs

        # use the temporary table to generate the table on the fly, as a list of EvColumns
        self.table = []
        for col in table:
            if isinstance(col, EvColumn):
                self.add_column(col, **kwargs)
            elif isinstance(col, (list, tuple)):
                self.table.append(EvColumn(*col, **kwargs))
            else:
                raise RuntimeError(
                    "EvTable 'table' kwarg must be a list of EvColumns or a list-of-lists of"
                    f" strings. Found {type(col)}."
                )

        # self.table = [EvColumn(*col, **kwargs) for col in table]

        # this is the actual working table
        self.worktable = None

        # balance the table
        # self._balance()

    def _cellborders(self, ix, iy, nx, ny, **kwargs):
        """
        Adds borders to the table by adjusting the input kwarg to
        instruct cells to build a border in the right positions.

        Args:
            ix (int): x index positions in table.
            iy (int): y index positions in table.
            nx (int): x size of table.
            ny (int): y size of table.

        Keyword Args:
            Keywords as per `EvTable.__init__`.

        Returns:
            table (str): string with the correct borders.

        Notes:
            A copy of the kwarg is returned to the cell. This is method
            is called by self._borders.

        """

        ret = kwargs.copy()

        # handle the various border modes
        border = self.border
        header = self.header

        bwidth = self.border_width
        headchar = self.header_line_char

        def corners(ret):
            """Handle corners of table"""
            if ix == 0 and iy == 0:
                ret["corner_top_left_char"] = self.corner_top_left_char
            if ix == nx and iy == 0:
                ret["corner_top_right_char"] = self.corner_top_right_char
            if ix == 0 and iy == ny:
                ret["corner_bottom_left_char"] = self.corner_bottom_left_char
            if ix == nx and iy == ny:
                ret["corner_bottom_right_char"] = self.corner_bottom_right_char
            return ret

        def left_edge(ret):
            """add vertical border along left table edge"""
            if ix == 0:
                ret["border_left"] = bwidth
                # ret["trim_horizontal"] = bwidth
            return ret

        def top_edge(ret):
            """add border along top table edge"""
            if iy == 0:
                ret["border_top"] = bwidth
                # ret["trim_vertical"] = bwidth
            return ret

        def right_edge(ret):
            """add vertical border along right table edge"""
            if ix == nx:  # and 0 < iy < ny:
                ret["border_right"] = bwidth
                # ret["trim_horizontal"] = 0
            return ret

        def bottom_edge(ret):
            """add border along bottom table edge"""
            if iy == ny:
                ret["border_bottom"] = bwidth
                # ret["trim_vertical"] = bwidth
            return ret

        def cols(ret):
            """Adding vertical borders inside the table"""
            if 0 <= ix < nx:
                ret["border_right"] = bwidth
            return ret

        def rows(ret):
            """Adding horizontal borders inside the table"""
            if 0 <= iy < ny:
                ret["border_bottom"] = bwidth
            return ret

        def head(ret):
            """Add header underline"""
            if iy == 0:
                # put different bottom line for header
                ret["border_bottom"] = bwidth
                ret["border_bottom_char"] = headchar
            return ret

        # use the helper functions to define various
        # table "styles"

        if border in ("table", "tablecols", "cells"):
            ret = bottom_edge(right_edge(top_edge(left_edge(corners(ret)))))
        if border in ("cols", "tablecols", "cells"):
            ret = cols(right_edge(left_edge(ret)))
        if border in "incols":
            ret = cols(ret)
        if border in ("rows", "cells"):
            ret = rows(bottom_edge(top_edge(ret)))
        if header and border not in ("none", None):
            ret = head(ret)

        return ret

    def _borders(self):
        """
        Add borders to table. This is called from self._balance.
        """
        nx, ny = self.ncols - 1, self.nrows - 1
        options = self.options
        for ix, col in enumerate(self.worktable):
            for iy, cell in enumerate(col):
                col.reformat_cell(iy, **self._cellborders(ix, iy, nx, ny, **options))

    def _balance(self):
        """
        Balance the table. This means to make sure
        all cells on the same row have the same height,
        that all columns have the same number of rows
        and that the table fits within the given width.
        """

        # we make all modifications on a working copy of the
        # actual table. This allows us to add columns/rows
        # and re-balance over and over without issue.
        self.worktable = deepcopy(self.table)
        #        self._borders()
        #        return
        options = copy(self.options)

        # balance number of rows to make a rectangular table
        # column by column
        ncols = len(self.worktable)
        nrows = [len(col) for col in self.worktable]
        nrowmax = max(nrows) if nrows else 0
        for icol, nrow in enumerate(nrows):
            self.worktable[icol].reformat(**options)
            if nrow < nrowmax:
                # add more rows to too-short columns
                empty_rows = ["" for _ in range(nrowmax - nrow)]
                self.worktable[icol].add_rows(*empty_rows)
        self.ncols = ncols
        self.nrows = nrowmax

        # add borders - these add to the width/height, so we must do this before calculating
        # width/height
        self._borders()

        # equalize widths within each column
        cwidths = [max(cell.get_width() for cell in col) for col in self.worktable]

        if self.width or self.maxwidth and self.maxwidth < sum(cwidths):
            # we set a table width. Horizontal cells will be evenly distributed and
            # expand vertically as needed (unless self.height is set, see below)

            # use fixed width, or set to maxwidth
            width = self.width if self.width else self.maxwidth

            if ncols:
                # get minimum possible cell widths for each row
                cwidths_min = [max(cell.get_min_width() for cell in col) for col in self.worktable]
                cwmin = sum(cwidths_min)

                # get which cols have separately set widths - these should be locked
                # note that we need to remove cwidths_min for each lock to avoid counting
                # it twice (in cwmin and in locked_cols)
                locked_cols = {
                    icol: col.options["width"] - cwidths_min[icol]
                    for icol, col in enumerate(self.worktable)
                    if "width" in col.options
                }
                locked_width = sum(locked_cols.values())

                excess = width - cwmin - locked_width

                if len(locked_cols) >= ncols and excess:
                    # we can't adjust the width at all - all columns are locked
                    raise Exception(
                        "Cannot balance table to width %s - "
                        "all columns have a set, fixed width summing to %s!"
                        % (self.width, sum(cwidths))
                    )

                if excess < 0:
                    # the locked cols makes it impossible
                    raise Exception(
                        "Cannot shrink table width to %s. "
                        "Minimum size (and/or fixed-width columns) "
                        "sets minimum at %s." % (self.width, cwmin + locked_width)
                    )

                if self.evenwidth:
                    # make each column of equal width
                    # use cwidths as a work-array to track weights
                    cwidths = copy(cwidths_min)
                    correction = 0
                    while correction < excess:
                        # flood-fill the minimum table starting with the smallest columns
                        ci = cwidths.index(min(cwidths))
                        if ci in locked_cols:
                            # locked column, make sure it's not picked again
                            cwidths[ci] += 9999
                            cwidths_min[ci] = locked_cols[ci]
                        else:
                            cwidths_min[ci] += 1
                            correction += 1
                    cwidths = cwidths_min
                else:
                    # make each column expand more proportional to their data size
                    # we use cwidth as a work-array to track weights
                    correction = 0
                    while correction < excess:
                        # fill wider columns first
                        ci = cwidths.index(max(cwidths))
                        if ci in locked_cols:
                            # locked column, make sure it's not picked again
                            cwidths[ci] -= 9999
                            cwidths_min[ci] = locked_cols[ci]
                        else:
                            cwidths_min[ci] += 1
                            correction += 1
                            # give a just changed col less prio next run
                            cwidths[ci] -= 3
                    cwidths = cwidths_min

        # reformat worktable (for width align)
        for ix, col in enumerate(self.worktable):
            try:
                col.reformat(width=cwidths[ix], **options)
            except Exception:
                raise

        # equalize heights for each row (we must do this here, since it may have changed to fit new
        # widths)
        cheights = [
            max(cell.get_height() for cell in (col[iy] for col in self.worktable))
            for iy in range(nrowmax)
        ]

        if self.height:
            # if we are fixing the table height, it means cells must crop text instead of resizing.
            if nrowmax:
                # get minimum possible cell heights for each column
                cheights_min = [
                    max(cell.get_min_height() for cell in (col[iy] for col in self.worktable))
                    for iy in range(nrowmax)
                ]
                chmin = sum(cheights_min)

                # get which cols have separately set heights - these should be locked
                # note that we need to remove cheights_min for each lock to avoid counting
                # it twice (in chmin and in locked_cols)
                locked_cols = {
                    icol: col.options["height"] - cheights_min[icol]
                    for icol, col in enumerate(self.worktable)
                    if "height" in col.options
                }
                locked_height = sum(locked_cols.values())

                excess = self.height - chmin - locked_height

                if chmin > self.height:
                    # we cannot shrink any more
                    raise Exception(
                        "Cannot shrink table height to %s. Minimum "
                        "size (and/or fixed-height rows) sets minimum at %s."
                        % (self.height, chmin + locked_height)
                    )

                # Add all the excess at the end of the table
                # Note: Older solutions tried to balance individual
                # rows' vsize. This could lead to empty rows that
                # looked like a bug. This solution instead
                # adds empty rows at the end which is less sophisticated
                # but much more visually consistent.
                cheights_min[-1] += excess
                cheights = cheights_min

                # we must tell cells to crop instead of expanding
            options["enforce_size"] = True

        # reformat table (for vertical align)
        for ix, col in enumerate(self.worktable):
            for iy, cell in enumerate(col):
                try:
                    col.reformat_cell(iy, height=cheights[iy], **options)
                except Exception as e:
                    msg = "ix=%s, iy=%s, height=%s: %s" % (ix, iy, cheights[iy], e.message)
                    raise Exception("Error in vertical align:\n %s" % msg)

        # calculate actual table width/height in characters
        self.cwidth = sum(cwidths)
        self.cheight = sum(cheights)

    def _generate_lines(self):
        """
        Generates lines across all columns
        (each cell may contain multiple lines)
        This will also balance the table.
        """
        self._balance()
        for iy in range(self.nrows):
            cell_row = [col[iy] for col in self.worktable]
            # this produces a list of lists, each of equal length
            cell_data = [cell.get() for cell in cell_row]
            cell_height = min(len(lines) for lines in cell_data)
            for iline in range(cell_height):
                yield ANSIString("").join(_to_ansi(celldata[iline] for celldata in cell_data))

    def add_header(self, *args, **kwargs):
        """
        Add header to table. This is a number of texts to be put at
        the top of the table. They will replace an existing header.

        Args:
            args (str): These strings will be used as the header texts.

        Keyword Args:
            Same keywords as per `EvTable.__init__`. Will be applied
            to the new header's cells.

        """
        self.header = True
        self.add_row(ypos=0, *args, **kwargs)

    def add_column(self, *args, **kwargs):
        """
        Add a column to table. If there are more rows in new column
        than there are rows in the current table, the table will
        expand with empty rows in the other columns. If too few, the
        new column with get new empty rows. All filling rows are added
        to the end.

        Args:
            args (`EvColumn` or multiple strings): Either a single EvColumn instance or
                a number of data string arguments to be used to create a new column.
            header (str, optional): The header text for the column
            xpos (int, optional): Index position in table *before* which
               to input new column. If not given, column will be added to the end
               of the table. Uses Python indexing (so first column is `xpos=0`)

        Keyword Args:
            Other keywords as per `Cell.__init__`.

        """
        # this will replace default options with new ones without changing default
        options = dict(list(self.options.items()) + list(kwargs.items()))

        xpos = kwargs.get("xpos", None)

        if args and isinstance(args[0], EvColumn):
            column = args[0]
            column.reformat(**kwargs)
        else:
            column = EvColumn(*args, **options)
        wtable = self.ncols
        htable = self.nrows

        header = kwargs.get("header", None)
        if header:
            column.add_rows(str(header), ypos=0, **options)
            self.header = True
        elif self.header:
            # we have a header already. Offset
            column.add_rows("", ypos=0, **options)

        # Calculate whether the new column needs to expand to the
        # current table size, or if the table needs to expand to
        # the column size.
        # This needs to happen after the header rows have already been
        # added to the column in order for the size calculations to match.
        excess = len(column) - htable
        if excess > 0:
            # we need to add new rows to table
            for col in self.table:
                empty_rows = ["" for _ in range(excess)]
                col.add_rows(*empty_rows, **options)
            self.nrows += excess
        elif excess < 0:
            # we need to add new rows to new column
            empty_rows = ["" for _ in range(abs(excess))]
            column.add_rows(*empty_rows, **options)

        if xpos is None or xpos > wtable - 1:
            # add to the end
            self.table.append(column)
        else:
            # insert column
            xpos = min(wtable - 1, max(0, int(xpos)))
            self.table.insert(xpos, column)
        self.ncols += 1

    def add_row(self, *args, **kwargs):
        """
        Add a row to table (not a header). If there are more cells in
        the given row than there are cells in the current table the
        table will be expanded with empty columns to match. These will
        be added to the end of the table. In the same way, adding a
        line with too few cells will lead to the last ones getting
        padded.

        Args:
            args (str): Any number of string argumnets to use as the
                data in the row (one cell per argument).
            ypos (int, optional): Index position in table before which to
                 input new row. If not given, will be added to the end of the table.
                 Uses Python indexing (so first row is `ypos=0`)

        Keyword Args:
            Other keywords are as per `EvCell.__init__`.

        """
        # this will replace default options with new ones without changing default
        row = list(args)
        options = dict(list(self.options.items()) + list(kwargs.items()))

        ypos = kwargs.get("ypos", None)
        wtable = self.ncols
        htable = self.nrows
        excess = len(row) - wtable

        if excess > 0:
            # we need to add new empty columns to table
            empty_rows = ["" for _ in range(htable)]
            self.table.extend([EvColumn(*empty_rows, **options) for _ in range(excess)])
        elif excess < 0:
            # we need to add more cells to row
            row.extend(["" for _ in range(abs(excess))])
        self.ncols = len(self.table)

        if ypos is None or ypos > htable - 1:
            # add new row to the end
            for icol, col in enumerate(self.table):
                col.add_rows(row[icol], **options)
        else:
            # insert row elsewhere
            ypos = min(htable - 1, max(0, int(ypos)))
            for icol, col in enumerate(self.table):
                col.add_rows(row[icol], ypos=ypos, **options)
        self.nrows += 1
        # self._balance()

    def reformat(self, **kwargs):
        """
        Force a re-shape of the entire table.

        Keyword Args:
            Table options as per `EvTable.__init__`.

        """
        self.width = kwargs.pop("width", self.width)
        self.height = kwargs.pop("height", self.height)
        for key, value in kwargs.items():
            setattr(self, key, value)

        hchar = kwargs.pop("header_line_char", self.header_line_char)

        # border settings are also passed on into EvCells (so kwargs.get, not kwargs.pop)
        self.header_line_char = hchar[0] if hchar else self.header_line_char
        self.border_width = kwargs.get("border_width", self.border_width)
        self.corner_char = kwargs.get("corner_char", self.corner_char)
        self.header_line_char = kwargs.get("header_line_char", self.header_line_char)

        self.corner_top_left_char = _to_ansi(kwargs.pop("corner_top_left_char", self.corner_char))
        self.corner_top_right_char = _to_ansi(kwargs.pop("corner_top_right_char", self.corner_char))
        self.corner_bottom_left_char = _to_ansi(
            kwargs.pop("corner_bottom_left_char", self.corner_char)
        )
        self.corner_bottom_right_char = _to_ansi(
            kwargs.pop("corner_bottom_right_char", self.corner_char)
        )

        self.options.update(kwargs)

    def reformat_column(self, index, **kwargs):
        """
        Sends custom options to a specific column in the table.

        Args:
            index (int): Which column to reformat. The column index is
                given from 0 to Ncolumns-1.

        Keyword Args:
            Column options as per `EvCell.__init__`.

        Raises:
            Exception: if an invalid index is found.

        """
        if index > len(self.table):
            raise Exception("Not a valid column index")
        # we update the columns' options which means eventual width/height
        # will be 'locked in' and withstand auto-balancing width/height from the table later
        self.table[index].options.update(kwargs)
        self.table[index].reformat(**kwargs)

    def get(self):
        """
        Return lines of table as a list.

        Returns:
            table_lines (list): The lines of the table, in order.

        """
        return [line for line in self._generate_lines()]

    def __str__(self):
        """print table (this also balances it)"""
        # h = "12345678901234567890123456789012345678901234567890123456789012345678901234567890"
        return str(str(ANSIString("\n").join([line for line in self._generate_lines()])))



"""
Roleplaying base system for Evennia

Contribution - Griatch, 2015

This module contains the ContribRPObject, ContribRPRoom and
ContribRPCharacter typeclasses.  If you inherit your
objects/rooms/character from these (or make them the defaults) from
these you will get the following features:

- Objects/Rooms will get the ability to have poses and will report
  the poses of items inside them (the latter most useful for Rooms).
- Characters will get poses and also sdescs (short descriptions)
that will be used instead of their keys. They will gain commands
for managing recognition (custom sdesc-replacement), masking
themselves as well as an advanced free-form emote command.

In more detail, This RP base system introduces the following features
to a game, common to many RP-centric games:

- emote system using director stance emoting (names/sdescs).
    This uses a customizable replacement noun (/me, @ etc) to
    represent you in the emote. You can use /sdesc, /nick, /key or
    /alias to reference objects in the room. You can use any
    number of sdesc sub-parts to differentiate a local sdesc, or
    use /1-sdesc etc to differentiate them. The emote also
    identifies nested says and separates case.
- sdesc obscuration of real character names for use in emotes
    and in any referencing such as object.search().  This relies
    on an SdescHandler `sdesc` being set on the Character and
    makes use of a custom Character.get_display_name hook. If
    sdesc is not set, the character's `key` is used instead. This
    is particularly used in the emoting system.
- recog system to assign your own nicknames to characters, can then
    be used for referencing. The user may recog a user and assign
    any personal nick to them. This will be shown in descriptions
    and used to reference them. This is making use of the nick
    functionality of Evennia.
- masks to hide your identity (using a simple lock).
- pose system to set room-persistent poses, visible in room
    descriptions and when looking at the person/object.  This is a
    simple Attribute that modifies how the characters is viewed when
    in a room as sdesc + pose.
- in-emote says, including seamless integration with language
    obscuration routine (such as contrib/rpg/rplanguage.py)

Installation:

Add `RPSystemCmdSet` from this module to your CharacterCmdSet:

```python
# mygame/commands/default_cmdsets.py

# ...

from evennia.contrib.rpg.rpsystem.rpsystem import RPSystemCmdSet  <---

class CharacterCmdSet(default_cmds.CharacterCmdset):
    # ...
    def at_cmdset_creation(self):
        # ...
        self.add(RPSystemCmdSet())  # <---

```

You also need to make your Characters/Objects/Rooms inherit from
the typeclasses in this module:

```python
# in mygame/typeclasses/characters.py

from evennia.contrib.rpg.rpsystem.rpsystem import ContribRPCharacter

class Character(ContribRPCharacter):
    # ...

```

```python
# in mygame/typeclasses/objects.py

from evennia.contrib.rpg.rpsystem.rpsystem import ContribRPObject

class Object(ContribRPObject):
    # ...

```

```python
# in mygame/typeclasses/rooms.py

from evennia.contrib.rpg.rpsystem.rpsystem import ContribRPRoom

class Room(ContribRPRoom):
    # ...

```

Examples:

> look
Tavern
The tavern is full of nice people

*A tall man* is standing by the bar.

Above is an example of a player with an sdesc "a tall man". It is also
an example of a static *pose*: The "standing by the bar" has been set
by the player of the tall man, so that people looking at him can tell
at a glance what is going on.

> emote /me looks at /Tall and says "Hello!"

I see:
    Griatch looks at Tall man and says "Hello".
Tall man (assuming his name is Tom) sees:
    The godlike figure looks at Tom and says "Hello".

Note that by default, the case of the tag matters, so `/tall` will
lead to 'tall man' while `/Tall` will become 'Tall man' and /TALL
becomes /TALL MAN. If you don't want this behavior, you can pass
case_sensitive=False to the `send_emote` function.

Extra Installation Instructions:

1. In typeclasses/character.py:
   Import the `ContribRPCharacter` class:
       `from evennia.contrib.rpg.rpsystem.rpsystem import ContribRPCharacter`
   Inherit ContribRPCharacter:
       Change "class Character(DefaultCharacter):" to
       `class Character(ContribRPCharacter):`
   If you have any overriden calls in `at_object_creation(self)`:
       Add `super().at_object_creation()` as the top line.
2. In `typeclasses/rooms.py`:
       Import the `ContribRPRoom` class:
       `from evennia.contrib.rpg.rpsystem.rpsystem import ContribRPRoom`
   Inherit `ContribRPRoom`:
       Change `class Room(DefaultRoom):` to
       `class Room(ContribRPRoom):`
3. In `typeclasses/objects.py`
       Import the `ContribRPObject` class:
       `from evennia.contrib.rpg.rpsystem.rpsystem import ContribRPObject`
   Inherit `ContribRPObject`:
       Change `class Object(DefaultObject):` to
       `class Object(ContribRPObject):`
4. Reload the server (`reload` or from console: "evennia reload")
5. Force typeclass updates as required. Example for your character:
       `type/reset/force me = typeclasses.characters.Character`

"""
import re
from collections import defaultdict
from string import punctuation

import inflect
from django.conf import settings

from evennia.commands.cmdset import CmdSet
from evennia.commands.command import Command
from evennia.objects.models import ObjectDB
from evennia.objects.objects import DefaultCharacter, DefaultObject
from evennia.utils import ansi, logger
from evennia.utils.utils import (
    iter_to_str,
    lazy_property,
    make_iter,
    variable_from_module,
)

_INFLECT = inflect.engine()

_AT_SEARCH_RESULT = variable_from_module(*settings.SEARCH_AT_RESULT.rsplit(".", 1))
# ------------------------------------------------------------
# Emote parser
# ------------------------------------------------------------

# Settings

# The prefix is the (single-character) symbol used to find the start
# of a object reference, such as /tall (note that
# the system will understand multi-word references like '/a tall man' too).
_PREFIX = getattr(settings, "RPSYSTEM_EMOTE_PREFIX", "/")

# The num_sep is the (single-character) symbol used to separate the
# sdesc from the number when  trying to separate identical sdescs from
# one another. This is the same syntax used in the rest of Evennia, so
# by default, multiple "tall" can be separated by entering 1-tall,
# 2-tall etc.
_NUM_SEP = "-"

# Texts

_EMOTE_NOMATCH_ERROR = """|RNo match for |r{ref}|R.|n"""

_EMOTE_MULTIMATCH_ERROR = """|RMultiple possibilities for {ref}:
    |r{reflist}|n"""

_RE_FLAGS = re.MULTILINE + re.IGNORECASE + re.UNICODE

_RE_PREFIX = re.compile(rf"^{_PREFIX}", re.UNICODE)

# This regex will return groups (num, word), where num is an optional counter to
# separate multimatches from one another and word is the first word in the
# marker. So entering "/tall man" will return groups ("", "tall")
# and "/2-tall man" will return groups ("2", "tall").
# the negative lookbehind for [:/] is to avoid http:// urls being detected as a /sdesc
_RE_OBJ_REF_START = re.compile(rf"(?<![:/]){_PREFIX}(?:([0-9]+){_NUM_SEP})*(\w+)", _RE_FLAGS)

_RE_LEFT_BRACKETS = re.compile(r"\{+", _RE_FLAGS)
_RE_RIGHT_BRACKETS = re.compile(r"\}+", _RE_FLAGS)
# Reference markers are used internally when distributing the emote to
# all that can see it. They are never seen by players and are on the form {#dbref<char>}
# with the <char> indicating case of the original reference query (like ^ for uppercase)
_RE_REF = re.compile(r"\{+\#([0-9]+[\^\~tv]{0,1})\}+")

# This regex is used to quickly reference one self in an emote.
_RE_SELF_REF = re.compile(r"(/me|@)(?=\W+)", _RE_FLAGS)

# regex for non-alphanumberic end of a string
_RE_CHAREND = re.compile(r"\W+$", _RE_FLAGS)

# reference markers for language
_RE_REF_LANG = re.compile(r"\{+\##([0-9]+)\}+")
# language says in the emote are on the form "..." or langname"..." (no spaces).
# this regex returns in groups (langname, say), where langname can be empty.
_RE_LANGUAGE = re.compile(r"(?:(\w+))*(\".+?\")")


# the emote parser works in two steps:
#  1) convert the incoming emote into an intermediary
#     form with all object references mapped to ids.
#  2) for every person seeing the emote, parse this
#     intermediary form into the one valid for that char.


class EmoteError(Exception):
    pass


class SdescError(Exception):
    pass


class RecogError(Exception):
    pass


class LanguageError(Exception):
    pass


def _get_case_ref(string):
    """
    Helper function which parses capitalization and
    returns the appropriate case-ref character for emotes.
    """
    # default to retaining the original case
    case = "~"
    # internal flags for the case used for the original /query
    # - t for titled input (like /Name)
    # - ^ for all upercase input (like /NAME)
    # - v for lower-case input (like /name)
    # - ~ for mixed case input (like /nAmE)
    if string.istitle():
        case = "t"
    elif string.isupper():
        case = "^"
    elif string.islower():
        case = "v"

    return case


# emoting mechanisms
def parse_language(speaker, emote):
    """
    Parse the emote for language. This is
    used with a plugin for handling languages.

    Args:
        speaker (Object): The object speaking.
        emote (str): An emote possibly containing
            language references.

    Returns:
        (emote, mapping) (tuple): A tuple where the
            `emote` is the emote string with all says
            (including quotes) replaced with reference
            markers on the form {##n} where n is a running
            number. The `mapping` is a dictionary between
            the markers and a tuple (langname, saytext), where
            langname can be None.
    Raises:
        evennia.contrib.rpg.rpsystem.LanguageError: If an invalid language was
        specified.

    Notes:
        Note that no errors are raised if the wrong language identifier
        is given.
        This data, together with the identity of the speaker, is
        intended to be used by the "listener" later, since with this
        information the language skill of the speaker can be offset to
        the language skill of the listener to determine how much
        information is actually conveyed.

    """
    # escape mapping syntax on the form {##id} if it exists already in emote,
    # if so it is replaced with just "id".
    emote = _RE_REF_LANG.sub(r"\1", emote)

    errors = []
    mapping = {}
    for imatch, say_match in enumerate(reversed(list(_RE_LANGUAGE.finditer(emote)))):
        # process matches backwards to be able to replace
        # in-place without messing up indexes for future matches
        # note that saytext includes surrounding "...".
        langname, saytext = say_match.groups()
        istart, iend = say_match.start(), say_match.end()
        # the key is simply the running match in the emote
        key = f"##{imatch}"
        # replace say with ref markers in emote
        emote = "{start}{{{key}}}{end}".format(start=emote[:istart], key=key, end=emote[iend:])
        mapping[key] = (langname, saytext)

    if errors:
        # catch errors and report
        raise LanguageError("\n".join(errors))

    # at this point all says have been replaced with {##nn} markers
    # and mapping maps 1:1 to this.
    return emote, mapping


def parse_sdescs_and_recogs(
    sender, candidates, string, search_mode=False, case_sensitive=True, fallback=None
):
    """
    Read a raw emote and parse it into an intermediary
    format for distributing to all observers.

    Args:
        sender (Object): The object sending the emote. This object's
            recog data will be considered in the parsing.
        candidates (iterable): A list of objects valid for referencing
            in the emote.
    string (str): The string (like an emote) we want to analyze for keywords.
    search_mode (bool, optional): If `True`, the "emote" is a query string
        we want to analyze. If so, the return value is changed.
    case_sensitive (bool, optional): If set, the case of /refs matter, so that
        /tall will come out as 'tall man' while /Tall will become 'Tall man'.
        This allows for more grammatically correct emotes at the cost of being
        a little more to learn for players. If disabled, the original sdesc case
        is always kept and are inserted as-is.
    fallback (string, optional): If set, any references that don't match a target
        will be replaced with the fallback string. If `None` (default), the
        parsing will fail and give a warning about the missing reference.

    Returns:
        (emote, mapping) (tuple): If `search_mode` is `False`
            (default), a tuple where the emote is the emote string, with
            all references replaced with internal-representation {#dbref}
            markers and mapping is a dictionary `{"#dbref":obj, ...}`.
        result (list): If `search_mode` is `True` we are
            performing a search query on `string`, looking for a specific
            object. A list with zero, one or more matches.

    Raises:
        EmoteException: For various ref-matching errors.

    Notes:
        The parser analyzes and should understand the following
        _PREFIX-tagged structures in the emote:
        - self-reference (/me)
        - recogs (any part of it) stored on emoter, matching obj in `candidates`.
        - sdesc (any part of it) from any obj in `candidates`.
        - N-sdesc, N-recog separating multi-matches (1-tall, 2-tall)
        - says, "..." are

    """
    # build a list of candidates with all possible referrable names
    # include 'me' keyword for self-ref
    candidate_map = []
    for obj in candidates:
        # check if sender has any recogs for obj and add
        if hasattr(sender, "recog"):
            if recog := sender.recog.get(obj):
                candidate_map.append((obj, recog))
        # check if obj has an sdesc and add
        if hasattr(obj, "sdesc"):
            candidate_map.append((obj, obj.sdesc.get()))
        # if no sdesc, include key plus aliases instead
        else:
            candidate_map.append((obj, obj.key))
        candidate_map.extend([(obj, alias) for alias in obj.aliases.all()])

    # escape mapping syntax on the form {#id} if it exists already in emote,
    # if so it is replaced with just "id".
    string = _RE_REF.sub(r"\1", string)
    # escape loose { } brackets since this will clash with formatting
    string = _RE_LEFT_BRACKETS.sub("{{", string)
    string = _RE_RIGHT_BRACKETS.sub("}}", string)

    # we now loop over all references and analyze them
    mapping = {}
    errors = []
    obj = None
    nmatches = 0
    # first, find and replace any self-refs
    for self_match in list(_RE_SELF_REF.finditer(string)):
        matched = self_match.group()
        case = _get_case_ref(matched.lstrip(_PREFIX)) if case_sensitive else ""
        key = f"#{sender.id}{case}"
        # replaced with ref
        string = _RE_SELF_REF.sub(f"{{{key}}}", string, count=1)
        mapping[key] = sender

    for marker_match in reversed(list(_RE_OBJ_REF_START.finditer(string))):
        # we scan backwards so we can replace in-situ without messing
        # up later occurrences. Given a marker match, query from
        # start index forward for all candidates.

        # first see if there is a number given (e.g. 1-tall)
        num_identifier, _ = marker_match.groups("")  # return "" if no match, rather than None
        match_index = marker_match.start()
        # split the emote string at the reference marker, to process everything after it
        head = string[:match_index]
        tail = string[match_index + 1 :]

        if search_mode:
            # match the candidates against the whole search string after the marker
            rquery = "".join(
                [
                    r"\b(" + re.escape(word.strip(punctuation)) + r").*"
                    for word in iter(tail.split())
                ]
            )
            matches = (
                (re.search(rquery, text, _RE_FLAGS), obj, text) for obj, text in candidate_map
            )
            # filter out any non-matching candidates
            bestmatches = [(obj, match.group()) for match, obj, text in matches if match]

        else:
            # to find the longest match, we start from the marker and lengthen the
            # match query one word at a time.
            word_list = []
            bestmatches = []
            # preserve punctuation when splitting
            tail = re.split("(\W)", tail)
            iend = 0
            for i, item in enumerate(tail):
                # don't add non-word characters to the search query
                if not item.isalpha():
                    continue
                word_list.append(item)
                rquery = "".join([r"\b(" + re.escape(word) + r").*" for word in word_list])
                # match candidates against the current set of words
                matches = (
                    (re.search(rquery, text, _RE_FLAGS), obj, text) for obj, text in candidate_map
                )
                matches = [(obj, match.group()) for match, obj, text in matches if match]
                if len(matches) == 0:
                    # no matches at this length, keep previous iteration as best
                    break
                # since this is the longest match so far, set latest match set as best matches
                bestmatches = matches
                # save current index as end point of matched text
                iend = i

            # save search string
            matched_text = "".join(tail[1:iend])
            # recombine remainder of emote back into a string
            tail = "".join(tail[iend + 1 :])

        nmatches = len(bestmatches)

        if not nmatches:
            # no matches
            obj = None
            nmatches = 0
        elif nmatches == 1:
            # an exact match.
            obj, match_str = bestmatches[0]
        elif all(bestmatches[0][0].id == obj.id for obj, text in bestmatches):
            # multi-match but all matches actually reference the same
            # obj (could happen with clashing recogs + sdescs)
            obj, match_str = bestmatches[0]
            nmatches = 1
        else:
            # multi-match.
            # was a numerical identifier given to help us separate the multi-match?
            inum = min(max(0, int(num_identifier) - 1), nmatches - 1) if num_identifier else None
            if inum is not None:
                # A valid inum is given. Use this to separate data.
                obj, match_str = bestmatches[inum]
                nmatches = 1
            else:
                # no identifier given - a real multimatch.
                obj = bestmatches

        if search_mode:
            # single-object search mode. Don't continue loop.
            break
        elif nmatches == 0:
            if fallback:
                # replace unmatched reference with the fallback string
                string = f"{head}{fallback}{tail}"
            else:
                errors.append(_EMOTE_NOMATCH_ERROR.format(ref=marker_match.group()))
        elif nmatches == 1:
            # a unique match - parse into intermediary representation
            case = _get_case_ref(marker_match.group()) if case_sensitive else ""
            # recombine emote with matched text replaced by ref
            key = f"#{obj.id}{case}"
            string = f"{head}{{{key}}}{tail}"
            mapping[key] = obj

        else:
            # multimatch error
            refname = marker_match.group()
            reflist = [
                "{num}{sep}{name} ({text}{key})".format(
                    num=inum + 1,
                    sep=_NUM_SEP,
                    name=_RE_PREFIX.sub("", refname),
                    text=text,
                    key=f" ({sender.key})" if sender == ob else "",
                )
                for inum, (ob, text) in enumerate(obj)
            ]
            errors.append(
                _EMOTE_MULTIMATCH_ERROR.format(
                    ref=marker_match.group(), reflist="\n    ".join(reflist)
                )
            )
    if search_mode:
        # return list of object(s) matching
        if nmatches == 0:
            return []
        elif nmatches == 1:
            return [obj]
        else:
            return [tup[0] for tup in obj]

    if errors:
        # make sure to not let errors through.
        raise EmoteError("\n".join(errors))

    # at this point all references have been replaced with {#xxx} markers and the mapping contains
    # a 1:1 mapping between those inline markers and objects.
    return string, mapping


def send_emote(sender, receivers, emote, msg_type="pose", anonymous_add="first", **kwargs):
    """
    Main access function for distribute an emote.

    Args:
        sender (Object): The one sending the emote.
        receivers (iterable): Receivers of the emote. These
            will also form the basis for which sdescs are
            'valid' to use in the emote.
        emote (str): The raw emote string as input by emoter.
        msg_type (str): The type of emote this is. "say" or "pose"
            for example. This is arbitrary and used for generating
            extra data for .msg(text) tuple.
        anonymous_add (str or None, optional): If `sender` is not
            self-referencing in the emote, this will auto-add
            `sender`'s data to the emote. Possible values are
            - None: No auto-add at anonymous emote
            - 'last': Add sender to the end of emote as [sender]
            - 'first': Prepend sender to start of emote.
    Kwargs:
        case_sensitive (bool): Defaults to True, but can be unset
            here. When enabled, /tall will lead to a lowercase
            'tall man' while /Tall will lead to 'Tall man' and
            /TALL will lead to 'TALL MAN'. If disabled, the sdesc's
            case will always be used, regardless of the /ref case used.
        any: Other kwargs will be passed on into the receiver's process_sdesc and
            process_recog methods, and can thus be used to customize those.

    """
    case_sensitive = kwargs.pop("case_sensitive", True)
    fallback = kwargs.pop("fallback", None)
    try:
        emote, obj_mapping = parse_sdescs_and_recogs(
            sender, receivers, emote, case_sensitive=case_sensitive, fallback=fallback
        )
        emote, language_mapping = parse_language(sender, emote)
    except (EmoteError, LanguageError) as err:
        # handle all error messages, don't hide actual coding errors
        sender.msg(str(err))
        return

    skey = f"#{sender.id}"

    # we escape the object mappings since we'll do the language ones first
    # (the text could have nested object mappings).
    emote = _RE_REF.sub(r"{{#\1}}", emote)
    # if anonymous_add is passed as a kwarg, collect and remove it from kwargs
    if "anonymous_add" in kwargs:
        anonymous_add = kwargs.pop("anonymous_add")
    # make sure to catch all possible self-refs
    self_refs = [f"{skey}{ref}" for ref in ("t", "^", "v", "~", "")]
    if anonymous_add and not any(1 for tag in obj_mapping if tag in self_refs):
        # no self-reference in the emote - add it
        if anonymous_add == "first":
            # add case flag for initial caps
            skey += "t"
            # don't put a space after the self-ref if it's a possessive emote
            femote = "{key}{emote}" if emote.startswith("'") else "{key} {emote}"
        else:
            # add it to the end
            femote = "{emote} [{key}]"
        emote = femote.format(key="{{" + skey + "}}", emote=emote)
        obj_mapping[skey] = sender

    # broadcast emote to everyone
    for receiver in receivers:
        # first handle the language mapping, which always produce different keys ##nn
        if hasattr(receiver, "process_language") and callable(receiver.process_language):
            receiver_lang_mapping = {
                key: receiver.process_language(saytext, sender, langname)
                for key, (langname, saytext) in language_mapping.items()
            }
        else:
            receiver_lang_mapping = {
                key: saytext for key, (langname, saytext) in language_mapping.items()
            }
        # map the language {##num} markers. This will convert the escaped sdesc markers on
        # the form {{#num}} to {#num} markers ready to sdesc-map in the next step.
        sendemote = emote.format_map(receiver_lang_mapping)

        # map the ref keys to sdescs
        receiver_sdesc_mapping = dict(
            (
                ref,
                obj.get_display_name(receiver, ref=ref, noid=True),
            )
            for ref, obj in obj_mapping.items()
        )

        # do the template replacement of the sdesc/recog {#num} markers
        receiver.msg(
            text=(sendemote.format_map(receiver_sdesc_mapping), {"type": msg_type}),
            from_obj=sender,
            **kwargs,
        )


# ------------------------------------------------------------
# Handlers for sdesc and recog
# ------------------------------------------------------------


class SdescHandler:
    """
    This Handler wraps all operations with sdescs. We
    need to use this since we do a lot preparations on
    sdescs when updating them, in order for them to be
    efficient to search for and query.

    The handler stores data in the following Attributes

        _sdesc   - a string
        _regex  - an empty dictionary

    """

    def __init__(self, obj):
        """
        Initialize the handler

        Args:
            obj (Object): The entity on which this handler is stored.

        """
        self.obj = obj
        self.sdesc = ""
        self._cache()

    def _cache(self):
        """
        Cache data from storage
        """
        self.sdesc = self.obj.attributes.get("_sdesc", default=self.obj.key)

    def add(self, sdesc, max_length=60):
        """
        Add a new sdesc to object, replacing the old one.

        Args:
            sdesc (str): The sdesc to set. This may be stripped
                of control sequences before setting.
            max_length (int, optional): The max limit of the sdesc.

        Returns:
            sdesc (str): The actually set sdesc.

        Raises:
            SdescError: If the sdesc is empty, can not be set or is
            longer than `max_length`.

        """
        # strip emote components from sdesc
        sdesc = _RE_REF.sub(
            r"\1",
            _RE_REF_LANG.sub(
                r"\1",
                _RE_SELF_REF.sub(r"", _RE_LANGUAGE.sub(r"", _RE_OBJ_REF_START.sub(r"", sdesc))),
            ),
        )

        # make an sdesc clean of ANSI codes
        cleaned_sdesc = ansi.strip_ansi(sdesc)

        if not cleaned_sdesc:
            raise SdescError("Short desc cannot be empty.")

        if len(cleaned_sdesc) > max_length:
            raise SdescError(
                "Short desc can max be {} chars long (was {} chars).".format(
                    max_length, len(cleaned_sdesc)
                )
            )

        # store to attributes
        self.obj.attributes.add("_sdesc", sdesc)
        # local caching
        self.sdesc = sdesc

        return sdesc

    def clear(self):
        """
        Clear sdesc.

        """
        self.obj.attributes.remove("_sdesc")

    def get(self):
        """
        Simple getter. The sdesc should never be allowed to
        be empty, but if it is we must fall back to the key.

        """
        return self.sdesc or self.obj.key


class RecogHandler:
    """
    This handler manages the recognition mapping
    of an Object.

    The handler stores data in Attributes as dictionaries of
    the following names:

        _recog_ref2recog
        _recog_obj2recog

    """

    def __init__(self, obj):
        """
        Initialize the handler

        Args:
            obj (Object): The entity on which this handler is stored.

        """
        self.obj = obj
        # mappings
        self.ref2recog = {}
        self.obj2recog = {}
        self._cache()

    def _cache(self):
        """
        Load data to handler cache
        """
        self.ref2recog = self.obj.attributes.get("_recog_ref2recog", default={})
        obj2recog = self.obj.attributes.get("_recog_obj2recog", default={})
        self.obj2recog = dict((obj, recog) for obj, recog in obj2recog.items() if obj)

    def add(self, obj, recog, max_length=60):
        """
        Assign a custom recog (nick) to the given object.

        Args:
            obj (Object): The object ot associate with the recog
                string. This is usually determined from the sdesc in the
                room by a call to parse_sdescs_and_recogs, but can also be
                given.
            recog (str): The replacement string to use with this object.
            max_length (int, optional): The max length of the recog string.

        Returns:
            recog (str): The (possibly cleaned up) recog string actually set.

        Raises:
            SdescError: When recog could not be set or sdesc longer
                than `max_length`.

        """
        if not obj.access(self.obj, "enable_recog", default=True):
            raise SdescError("This person is unrecognizeable.")

        # strip emote components from recog
        recog = _RE_REF.sub(
            r"\1",
            _RE_REF_LANG.sub(
                r"\1",
                _RE_SELF_REF.sub(r"", _RE_LANGUAGE.sub(r"", _RE_OBJ_REF_START.sub(r"", recog))),
            ),
        )

        # make an recog clean of ANSI codes
        cleaned_recog = ansi.strip_ansi(recog)

        if not cleaned_recog:
            raise SdescError("Recog string cannot be empty.")

        if len(cleaned_recog) > max_length:
            raise RecogError(
                "Recog string cannot be longer than {} chars (was {} chars)".format(
                    max_length, len(cleaned_recog)
                )
            )

        # mapping #dbref:obj
        key = f"#{obj.id}"
        self.obj.attributes.get("_recog_ref2recog", default={})[key] = recog
        self.obj.attributes.get("_recog_obj2recog", default={})[obj] = recog
        # local caching
        self.ref2recog[key] = recog
        self.obj2recog[obj] = recog
        return recog

    def get(self, obj):
        """
        Get recog replacement string, if one exists.

        Args:
            obj (Object): The object, whose sdesc to replace
        Returns:
            recog (str or None): The replacement string to use, or
                None if there is no recog for this object.

        Notes:
            This method will respect a "enable_recog" lock set on
            `obj` (True by default) in order to turn off recog
            mechanism. This is useful for adding masks/hoods etc.
        """
        if obj.access(self.obj, "enable_recog", default=True):
            # check an eventual recog_masked lock on the object
            # to avoid revealing masked characters. If lock
            # does not exist, pass automatically.
            return self.obj2recog.get(obj, None)
        else:
            # recog_mask lock not passed, disable recog
            return None

    def all(self):
        """
        Get a mapping of the recogs stored in handler.

        Returns:
            recogs (dict): A mapping of {recog: obj} stored in handler.

        """
        return {self.obj2recog[obj]: obj for obj in self.obj2recog.keys()}

    def remove(self, obj):
        """
        Clear recog for a given object.

        Args:
            obj (Object): The object for which to remove recog.
        """
        if obj in self.obj2recog:
            del self.obj.db._recog_obj2recog[obj]
            del self.obj.db._recog_ref2recog[f"#{obj.id}"]
        self._cache()


# ------------------------------------------------------------
# RP Commands
# ------------------------------------------------------------


class RPCommand(Command):
    "simple parent"

    def parse(self):
        "strip extra whitespace"
        self.args = self.args.strip()


class CmdEmote(RPCommand):  # replaces the main emote
    """
    Emote an action, allowing dynamic replacement of
    text in the emote.

    Usage:
      emote text

    Example:
      emote {prefix}me looks around.
      emote With a flurry {prefix}me attacks {prefix}tall man with his sword.
      emote "Hello", {prefix}me says.

    Describes an event in the world. This allows the use of {prefix}ref
    markers to replace with the short descriptions or recognized
    strings of objects in the same room. These will be translated to
    emotes to match each person seeing it. Use "..." for saying
    things and langcode"..." without spaces to say something in
    a different language.

    """

    key = "emote"
    aliases = [":"]
    locks = "cmd:all()"
    arg_regex = ""

    def get_help(self, caller, cmdset):
        return self.__doc__.format(prefix=_PREFIX)

    def func(self):
        "Perform the emote."
        if not self.args:
            self.caller.msg("What do you want to do?")
        else:
            # we also include ourselves here.
            emote = self.args
            targets = self.caller.location.contents
            if not emote.endswith((".", "?", "!", '"')):  # If emote is not punctuated or speech,
                emote += "."  # add a full-stop for good measure.
            send_emote(self.caller, targets, emote, anonymous_add="first")


class CmdSay(RPCommand):  # replaces standard say
    """
    speak as your character

    Usage:
      say <message>

    Talk to those in your current location.
    """

    key = "say"
    aliases = ['"', "'"]
    locks = "cmd:all()"
    arg_regex = ""

    def func(self):
        "Run the say command"

        caller = self.caller

        if not self.args:
            caller.msg("Say what?")
            return

        # calling the speech modifying hook
        speech = caller.at_pre_say(self.args)
        targets = self.caller.location.contents
        send_emote(self.caller, targets, speech, msg_type="say", anonymous_add=None)


class CmdSdesc(RPCommand):  # set/look at own sdesc
    """
    Assign yourself a short description (sdesc).

    Usage:
      sdesc <short description>
      sdesc         - view current sdesc
      sdesc clear   - remove sdesc

    Assigns a short description to yourself.

    """

    key = "sdesc"
    locks = "cmd:all()"

    def func(self):
        "Assign the sdesc"
        caller = self.caller
        if not self.args:
            sdesc = caller.sdesc.get()
            if not sdesc:
                caller.msg("You have no short description set.")
            else:
                caller.msg(f'Your short description is "{sdesc}".')
        elif self.args == "clear":
            ret = yield "Do you want to clear your sdesc? [Y]/n?"
            if ret.lower() in ("n", "no"):
                caller.msg("Aborted.")
            else:
                caller.sdesc.clear()
                caller.msg(f'Cleared sdesc, using name "{caller.key}".')
        else:
            # strip non-alfanum chars from end of sdesc
            sdesc = _RE_CHAREND.sub("", self.args)
            try:
                sdesc = caller.sdesc.add(sdesc)
            except SdescError as err:
                caller.msg(err)
                return
            except AttributeError:
                caller.msg(f"Cannot set sdesc on {caller.key}.")
                return
            caller.msg(f"{caller.key}'s sdesc was set to '{sdesc}'.")


class CmdPose(RPCommand):  # set current pose and default pose
    """
    Set a static pose

    Usage:
        pose <pose>
        pose default <pose>
        pose reset
        pose obj = <pose>
        pose default obj = <pose>
        pose reset obj =

    Examples:
        pose leans against the tree
        pose is talking to the barkeep.
        pose box = is sitting on the floor.

    Set a static pose. This is the end of a full sentence that starts
    with your sdesc. If no full stop is given, it will be added
    automatically. The default pose is the pose you get when using
    pose reset. Note that you can use sdescs/recogs to reference
    people in your pose, but these always appear as that person's
    sdesc in the emote, regardless of who is seeing it.

    """

    key = "pose"

    def parse(self):
        """
        Extract the "default" alternative to the pose.
        """
        args = self.args.strip()
        default = args.startswith("default")
        reset = args.startswith("reset")
        if default:
            args = re.sub(r"^default", "", args)
        if reset:
            args = re.sub(r"^reset", "", args)
        target = None
        if "=" in args:
            target, args = [part.strip() for part in args.split("=", 1)]

        self.target = target
        self.reset = reset
        self.default = default
        self.args = args.strip()

    def func(self):
        "Create the pose"
        caller = self.caller
        pose = self.args
        target = self.target
        if not pose and not self.reset:
            caller.msg("Usage: pose <pose-text> OR pose obj = <pose-text>")
            return

        if not pose.endswith((".", "?", "!", '"')):
            pose += "."
        if target:
            # affect something else
            target = caller.search(target)
            if not target:
                return
            if not target.access(caller, "edit"):
                caller.msg("You can't pose that.")
                return
        else:
            target = caller

        target_name = target.sdesc.get() if hasattr(target, "sdesc") else target.key
        if not target.attributes.has("pose"):
            caller.msg(f"{target_name} cannot be posed.")
            return

        # set the pose
        if self.reset:
            pose = target.db.pose_default
            target.db.pose = pose
        elif self.default:
            target.db.pose_default = pose
            caller.msg(f"Default pose is now '{target_name} {pose}'.")
            return
        else:
            # set the pose. We do one-time ref->sdesc mapping here.
            parsed, mapping = parse_sdescs_and_recogs(caller, caller.location.contents, pose)
            mapping = dict(
                (ref, obj.sdesc.get() if hasattr(obj, "sdesc") else obj.key)
                for ref, obj in mapping.items()
            )
            pose = parsed.format_map(mapping)

            if len(target_name) + len(pose) > 60:
                caller.msg(f"'{pose}' is too long.")
                return

            target.db.pose = pose

        caller.msg(f"Pose will read '{target_name} {pose}'.")


class CmdRecog(RPCommand):  # assign personal alias to object in room
    """
    Recognize another person in the same room.

    Usage:
      recog
      recog sdesc as alias
      forget alias

    Example:
        recog tall man as Griatch
        forget griatch

    This will assign a personal alias for a person, or forget said alias.
    Using the command without arguments will list all current recogs.

    """

    key = "recog"
    aliases = ["recognize", "forget"]

    def parse(self):
        "Parse for the sdesc as alias structure"
        self.sdesc, self.alias = "", ""
        if " as " in self.args:
            self.sdesc, self.alias = [part.strip() for part in self.args.split(" as ", 2)]
        elif self.args:
            # try to split by space instead
            try:
                self.sdesc, self.alias = [part.strip() for part in self.args.split(None, 1)]
            except ValueError:
                self.sdesc, self.alias = self.args.strip(), ""

    def func(self):
        "Assign the recog"
        caller = self.caller
        alias = self.alias.rstrip(".?!")
        sdesc = self.sdesc

        recog_mode = self.cmdstring != "forget" and alias and sdesc
        forget_mode = self.cmdstring == "forget" and sdesc
        list_mode = not self.args

        if not (recog_mode or forget_mode or list_mode):
            caller.msg("Usage: recog, recog <sdesc> as <alias> or forget <alias>")
            return

        if list_mode:
            # list all previously set recogs
            all_recogs = caller.recog.all()
            if not all_recogs:
                caller.msg(
                    "You recognize no-one. (Use 'recog <sdesc> as <alias>' to recognize people."
                )
            else:
                # note that we don't skip those failing enable_recog lock here,
                # because that would actually reveal more than we want.
                lst = "\n".join(
                    " {}  ({})".format(key, obj.sdesc.get() if hasattr(obj, "sdesc") else obj.key)
                    for key, obj in all_recogs.items()
                )
                caller.msg(
                    "Currently recognized (use 'recog <sdesc> as <alias>' to add "
                    f"new and 'forget <alias>' to remove):\n{lst}"
                )
            return

        prefixed_sdesc = sdesc if sdesc.startswith(_PREFIX) else _PREFIX + sdesc
        candidates = caller.location.contents
        matches = parse_sdescs_and_recogs(caller, candidates, prefixed_sdesc, search_mode=True)
        nmatches = len(matches)
        # handle 0 and >1 matches
        if nmatches == 0:
            caller.msg(_EMOTE_NOMATCH_ERROR.format(ref=sdesc))
        elif nmatches > 1:
            reflist = [
                "{num}{sep}{sdesc} ({recog}{key})".format(
                    num=inum + 1,
                    sep=_NUM_SEP,
                    sdesc=_RE_PREFIX.sub("", sdesc),
                    recog=caller.recog.get(obj) or "no recog",
                    key=f" ({caller.key})" if caller == obj else "",
                )
                for inum, obj in enumerate(matches)
            ]
            caller.msg(_EMOTE_MULTIMATCH_ERROR.format(ref=sdesc, reflist="\n    ".join(reflist)))

        else:
            # one single match
            obj = matches[0]
            if not obj.access(self.obj, "enable_recog", default=True):
                # don't apply recog if object doesn't allow it (e.g. by being masked).
                caller.msg("It's impossible to recognize them.")
                return
            if forget_mode:
                # remove existing recog
                caller.recog.remove(obj)
                caller.msg(
                    "You will now know them only as '{}'.".format(
                        obj.get_display_name(caller, noid=True)
                    )
                )
            else:
                # set recog
                sdesc = obj.sdesc.get() if hasattr(obj, "sdesc") else obj.key
                try:
                    alias = caller.recog.add(obj, alias)
                except RecogError as err:
                    caller.msg(err)
                    return
                caller.msg("You will now remember |w{}|n as |w{}|n.".format(sdesc, alias))


class CmdMask(RPCommand):
    """
    Wear a mask

    Usage:
        mask <new sdesc>
        unmask

    This will put on a mask to hide your identity. When wearing
    a mask, your sdesc will be replaced by the sdesc you pick and
    people's recognitions of you will be disabled.

    """

    key = "mask"
    aliases = ["unmask"]

    def func(self):
        caller = self.caller
        if self.cmdstring == "mask":
            # wear a mask
            if not self.args:
                caller.msg("Usage: (un)mask sdesc")
                return
            if caller.db.unmasked_sdesc:
                caller.msg("You are already wearing a mask.")
                return
            sdesc = _RE_CHAREND.sub("", self.args)
            sdesc = f"{sdesc} |H[masked]|n"
            if len(sdesc) > 60:
                caller.msg("Your masked sdesc is too long.")
                return
            caller.db.unmasked_sdesc = caller.sdesc.get()
            caller.locks.add("enable_recog:false()")
            caller.sdesc.add(sdesc)
            caller.msg(f"You wear a mask as '{sdesc}'.")
        else:
            # unmask
            old_sdesc = caller.db.unmasked_sdesc
            if not old_sdesc:
                caller.msg("You are not wearing a mask.")
                return
            del caller.db.unmasked_sdesc
            caller.locks.remove("enable_recog")
            caller.sdesc.add(old_sdesc)
            caller.msg(f"You remove your mask and are again '{old_sdesc}'.")


class RPSystemCmdSet(CmdSet):
    """
    Mix-in for adding rp-commands to default cmdset.
    """

    def at_cmdset_creation(self):
        self.add(CmdEmote())
        self.add(CmdSay())
        self.add(CmdSdesc())
        self.add(CmdPose())
        self.add(CmdRecog())
        self.add(CmdMask())


# ------------------------------------------------------------
# RP typeclasses
# ------------------------------------------------------------


class ContribRPObject(DefaultObject):
    """
    This class is meant as a mix-in or parent for objects in an
    rp-heavy game. It implements the base functionality for poses.
    """

    @lazy_property
    def sdesc(self):
        return SdescHandler(self)

    def at_object_creation(self):
        """
        Called at initial creation.
        """
        super().at_object_creation()

        # emoting/recog data
        self.db.pose = ""
        self.db.pose_default = "is here."
        self.db._sdesc = ""

    def get_search_result(
        self,
        searchdata,
        attribute_name=None,
        typeclass=None,
        candidates=None,
        exact=False,
        use_dbref=None,
        tags=None,
        **kwargs,
    ):
        """
        Override of the parent method for producing search results that understands sdescs.
        These are used in the main .search() method of the parent class.

        """
        # we also want to use the default search method
        search_obj = super().get_search_result
        is_builder = self.locks.check_lockstring(self, "perm(Builder)")

        if candidates:
            candidates = parse_sdescs_and_recogs(
                self, candidates, _PREFIX + searchdata, search_mode=True
            )
            results = []
            for candidate in candidates:
                # we search by candidate keys here; this allows full error
                # management and use of all kwargs - we will use searchdata
                # in eventual error reporting later (not their keys). Doing
                # it like this e.g. allows for use of the typeclass kwarg
                # limiter.
                results.extend([obj for obj in search_obj(candidate.key) if obj not in results])

            if not results and is_builder:
                # builders get a chance to search only by key+alias
                results = search_obj(searchdata)
        else:
            # global searches / #drefs end up here. Global searches are
            # only done in code, so is controlled, #dbrefs are turned off
            # for non-Builders.
            results = search_obj(searchdata)
        return results

    def get_posed_sdesc(self, sdesc, **kwargs):
        """
        Displays the object with its current pose string.

        Returns:
            pose (str): A string containing the object's sdesc and
                current or default pose.
        """

        # get the current pose, or default if no pose is set
        pose = self.db.pose or self.db.pose_default

        # return formatted string, or sdesc as fallback
        return f"{sdesc} {pose}" if pose else sdesc

    def get_display_name(self, looker, **kwargs):
        """
        Displays the name of the object in a viewer-aware manner.

        Args:
            looker (TypedObject): The object or account that is looking
                at/getting inforamtion for this object.

        Keyword Args:
            pose (bool): Include the pose (if available) in the return.
            ref (str): The reference marker found in string to replace.
                This is on the form #{num}{case}, like '#12^', where
                the number is a processing location in the string and the
                case symbol indicates the case of the original tag input
                - `t` - input was Titled, like /Tall
                - `^` - input was all uppercase, like /TALL
                - `v` - input was all lowercase, like /tall
                - `~` - input case should be kept, or was mixed-case
            noid (bool): Don't show DBREF even if viewer has control access.

        Returns:
            name (str): A string of the sdesc containing the name of the object,
                if this is defined. By default, included the DBREF if this user
                is privileged to control said object.

        """
        ref = kwargs.get("ref", "~")

        if looker == self:
            # always show your own key
            sdesc = self.key
        else:
            try:
                # get the sdesc looker should see
                sdesc = looker.get_sdesc(self, ref=ref)
            except AttributeError:
                # use own sdesc as a fallback
                sdesc = self.sdesc.get()

        # add dbref is looker has control access and `noid` is not set
        if self.access(looker, access_type="control") and not kwargs.get("noid", False):
            sdesc = f"{sdesc}(#{self.id})"

        return self.get_posed_sdesc(sdesc) if kwargs.get("pose", False) else sdesc

    def get_display_characters(self, looker, pose=True, **kwargs):
        """
        Get the ‘characters’ component of the object description. Called by return_appearance.
        """

        def _filter_visible(obj_list):
            return (obj for obj in obj_list if obj != looker and obj.access(looker, "view"))

        characters = _filter_visible(self.contents_get(content_type="character"))
        character_names = "\n".join(
            char.get_display_name(looker, pose=pose, **kwargs) for char in characters
        )

        return f"\n{character_names}" if character_names else ""

    def get_display_things(self, looker, pose=True, **kwargs):
        """
        Get the 'things' component of the object description. Called by `return_appearance`.

        Args:
            looker (Object): Object doing the looking.
            **kwargs: Arbitrary data for use when overriding.
        Returns:
            str: The things display data.

        """
        if not pose:
            # if poses aren't included, we can use the core version instead
            return super().get_display_things(looker, **kwargs)

        def _filter_visible(obj_list):
            return [obj for obj in obj_list if obj != looker and obj.access(looker, "view")]

        # sort and handle same-named things
        things = _filter_visible(self.contents_get(content_type="object"))

        posed_things = defaultdict(list)
        for thing in things:
            pose = thing.db.pose or thing.db.pose_default
            if not pose:
                pose = ""
            posed_things[pose].append(thing)

        display_strings = []

        for pose, thinglist in posed_things.items():
            grouped_things = defaultdict(list)
            for thing in thinglist:
                grouped_things[thing.get_display_name(looker, pose=False, **kwargs)].append(thing)

            thing_names = []
            for thingname, samethings in sorted(grouped_things.items()):
                nthings = len(samethings)
                thing = samethings[0]
                singular, plural = thing.get_numbered_name(nthings, looker, key=thingname)
                thing_names.append(singular if nthings == 1 else plural)
            thing_names = iter_to_str(thing_names)

            if pose:
                pose = _INFLECT.plural(pose) if nthings != 1 else pose
            grouped_names = f"{thing_names} {pose}"
            grouped_names = grouped_names[0].upper() + grouped_names[1:]
            display_strings.append(grouped_names)

        if not display_strings:
            return ""

        return "\n" + "\n".join(display_strings)


class ContribRPRoom(ContribRPObject):
    """
    Dummy inheritance for rooms.
    """

    pass


class ContribRPCharacter(DefaultCharacter, ContribRPObject):
    """
    This is a character class that has poses, sdesc and recog.
    """

    @lazy_property
    def recog(self):
        return RecogHandler(self)

    def get_display_name(self, looker, **kwargs):
        """
        Displays the name of the object in a viewer-aware manner.

        Args:
            looker (TypedObject): The object or account that is looking
                at/getting inforamtion for this object.

        Keyword Args:
            pose (bool): Include the pose (if available) in the return.
            ref (str): The reference marker found in string to replace.
                This is on the form #{num}{case}, like '#12^', where
                the number is a processing location in the string and the
                case symbol indicates the case of the original tag input
                - `t` - input was Titled, like /Tall
                - `^` - input was all uppercase, like /TALL
                - `v` - input was all lowercase, like /tall
                - `~` - input case should be kept, or was mixed-case
            noid (bool): Don't show DBREF even if viewer has control access.

        Returns:
            name (str): A string of the sdesc containing the name of the object,
                if this is defined. By default, included the DBREF if this user
                is privileged to control said object.

        Notes:
            The RPCharacter version adds additional processing to sdescs to make
            characters stand out from other objects.

        """
        ref = kwargs.get("ref", "~")

        if looker == self:
            # process your key as recog since you recognize yourself
            sdesc = self.process_recog(self.key, self)
        else:
            try:
                # get the sdesc looker should see, with formatting
                sdesc = looker.get_sdesc(self, process=True, ref=ref)
            except AttributeError:
                # use own sdesc as a fallback
                sdesc = self.sdesc.get()

        # add dbref is looker has control access and `noid` is not set
        if self.access(looker, access_type="control") and not kwargs.get("noid", False):
            sdesc = f"{sdesc}(#{self.id})"

        return self.get_posed_sdesc(sdesc) if kwargs.get("pose", False) else sdesc

    def at_object_creation(self):
        """
        Called at initial creation.
        """
        super().at_object_creation()

        self.db._sdesc = ""

        self.db._recog_ref2recog = {}
        self.db._recog_obj2recog = {}

        self.cmdset.add(RPSystemCmdSet, persistent=True)
        # initializing sdesc
        self.sdesc.add("A normal person")

    def at_pre_say(self, message, **kwargs):
        """
        Called before the object says or whispers anything, return modified message.

        Args:
            message (str): The suggested say/whisper text spoken by self.
        Keyword Args:
            whisper (bool): If True, this is a whisper rather than a say.

        """
        if kwargs.get("whisper"):
            return f'/Me whispers "{message}"'
        return f'/Me says, "{message}"'

    def get_sdesc(self, obj, process=False, **kwargs):
        """
        Single method to handle getting recogs with sdesc fallback in an
        aware manner, to allow separate processing of recogs from sdescs.
        Gets the sdesc or recog for obj from the view of self.

        Args:
            obj (Object): the object whose sdesc or recog is being gotten
        Keyword Args:
            process (bool): If True, the sdesc/recog is run through the
                appropriate process method for self - .process_sdesc or
                .process_recog
        """
        # always see own key
        if obj == self:
            recog = self.key
            sdesc = self.key
        else:
            # first check if we have a recog for this object
            recog = self.recog.get(obj)
            # set sdesc to recog, using sdesc as a fallback, or the object's key if no sdesc
            sdesc = recog or (hasattr(obj, "sdesc") and obj.sdesc.get()) or obj.key

        if process:
            # process the sdesc as a recog if a recog was found, else as an sdesc
            sdesc = (self.process_recog if recog else self.process_sdesc)(sdesc, obj, **kwargs)

        return sdesc

    def process_sdesc(self, sdesc, obj, **kwargs):
        """
        Allows to customize how your sdesc is displayed (primarily by
        changing colors).

        Args:
            sdesc (str): The sdesc to display.
            obj (Object): The object to which the adjoining sdesc
                belongs. If this object is equal to yourself, then
                you are viewing yourself (and sdesc is your key).
                This is not used by default.

        Kwargs:
            ref (str): The reference marker found in string to replace.
                This is on the form #{num}{case}, like '#12^', where
                the number is a processing location in the string and the
                case symbol indicates the case of the original tag input
                - `t` - input was Titled, like /Tall
                - `^` - input was all uppercase, like /TALL
                - `v` - input was all lowercase, like /tall
                - `~` - input case should be kept, or was mixed-case

        Returns:
            sdesc (str): The processed sdesc ready
                for display.

        """
        if not sdesc:
            return ""

        ref = kwargs.get("ref", "~")  # ~ to keep sdesc unchanged
        if "t" in ref:
            # we only want to capitalize the first letter if there are many words
            sdesc = sdesc.lower()
            sdesc = sdesc[0].upper() + sdesc[1:] if len(sdesc) > 1 else sdesc.upper()
        elif "^" in ref:
            sdesc = sdesc.upper()
        elif "v" in ref:
            sdesc = sdesc.lower()
        return f"|b{sdesc}|n"

    def process_recog(self, recog, obj, **kwargs):
        """
        Allows to customize how a recog string is displayed.

        Args:
            recog (str): The recog string. It has already been
                translated from the original sdesc at this point.
            obj (Object): The object the recog:ed string belongs to.
                This is not used by default.

        Returns:
            recog (str): The modified recog string.

        """
        if not recog:
            return ""

        return f"|m{recog}|n"

    def process_language(self, text, speaker, language, **kwargs):
        """
        Allows to process the spoken text, for example
        by obfuscating language based on your and the
        speaker's language skills. Also a good place to
        put coloring.

        Args:
            text (str): The text to process.
            speaker (Object): The object delivering the text.
            language (str): An identifier string for the language.

        Return:
            text (str): The optionally processed text.

        Notes:
            This is designed to work together with a string obfuscator
            such as the `obfuscate_language` or `obfuscate_whisper` in
            the evennia.contrib.rpg.rplanguage module.

        """
        return "{label}|w{text}|n".format(label=f"|W({language})" if language else "", text=text)
